"""
StructureFactorConstraints contains classes for all constraints related experimental static structure factor functions.

.. inheritance-diagram:: fullrmc.Constraints.StructureFactorConstraints
    :parts: 1
"""

# standard libraries imports
import itertools

# external libraries imports
import numpy as np
from pdbParser.Utilities.Database import is_element_property, get_element_property
from pdbParser.Utilities.Collection import get_normalized_weighting

# fullrmc imports
from fullrmc.Globals import INT_TYPE, FLOAT_TYPE, PI, PRECISION, LOGGER
from fullrmc.Core.Collection import is_number, is_integer, get_path
from fullrmc.Core.Constraint import Constraint, ExperimentalConstraint
from fullrmc.Core.pairs_histograms import multiple_pairs_histograms_coords, full_pairs_histograms_coords

class StructureFactorConstraint(ExperimentalConstraint):
    """
    It controls the Structure Factor noted as S(Q) and also called
    total-scattering structure function or Static Structure Factor. 
    S(Q) is a dimensionless quantity and the normalization is such 
    that the average value, :math:`<S(Q)>=1`.
    
    It is worth mentioning that S(Q) is nothing other than the normalized and 
    corrected diffraction pattern from all experimental artefacts powder.
    
    The computation of S(Q) is done through an inverse Sine Fourier transform  
    of the computed pair distribution function noted as G(r).
    
    .. math::
        
        S(Q) = 1+ \\frac{1}{Q} \\int_{0}^{\\infty} G(r) sin(Qr) dr
        
    From an atomistic model and histogram point of view, G(r) is computed as
    the following:
    
    .. math::
        
        G(r) = 4 \\pi r (\\rho_{r} - \\rho_{0})
             = 4 \\pi \\rho_{0} r (g(r)-1) 
             = \\frac{R(r)}{r} - 4 \\pi \\rho_{0}
    
    g(r) is calculated after binning all pair atomic distances into a 
    weighted histograms as the following:

    .. math::
        g(r) = \\sum \\limits_{i,j}^{N} w_{i,j} \\frac{\\rho_{i,j}(r)}{\\rho_{0}} 
             = \\sum \\limits_{i,j}^{N} w_{i,j} \\frac{n_{i,j}(r) / v(r)}{N_{i,j} / V} 
     
    Where:\n
    :math:`Q` is the momentum transfer. \n
    :math:`r` is the distance between two atoms. \n
    :math:`\\rho_{i,j}(r)` is the pair density function of atoms i and j. \n
    :math:`\\rho_{0}` is the  average number density of the system. \n
    :math:`w_{i,j}` is the relative weighting of atom types i and j. \n
    :math:`R(r)` is the radial distribution function (rdf). \n
    :math:`N` is the total number of atoms. \n
    :math:`V` is the volume of the system. \n
    :math:`n_{i,j}(r)` is the number of atoms i neighbouring j at a distance r. \n
    :math:`v(r)` is the annulus volume at distance r and of thickness dr. \n
    :math:`N_{i,j}` is the total number of atoms i and j in the system. \n

    
    :Parameters:
        #. engine (None, fullrmc.Engine): The constraint RMC engine.
        #. experimentalData (numpy.ndarray, string): The experimental data as numpy.ndarray or string path to load data using numpy.loadtxt.
        #. dataWeights (None, numpy.ndarray): A weights array of the same number of points of experimentalData used in the constraint's standard error computation.
           Therefore particular fitting emphasis can be put on different data points that might be considered as more or less
           important in order to get a reasonable and plausible modal.\n
           If None is given, all data points are considered of the same importance in the computation of the constraint's standard error.\n
           If numpy.ndarray is given, all weights must be positive and all zeros weighted data points won't contribute to the 
           total constraint's standard error. At least a single weight point is required to be non-zeros and the weights 
           array will be automatically scaled upon setting such as the the sum of all the weights is equal to the number of data points.       
        #. weighting (string): The elements weighting.
        #. rmin (None, number): The minimum distance value to compute G(r) histogram.
           If None is given, rmin is computed as :math:`2 \\pi / Q_{max}`.
        #. rmax (None, number): The maximum distance value to compute G(r) histogram.
           If None is given, rmax is computed as :math:`2 \\pi / dQ`.
        #. dr (None, number): The distance bin value to compute G(r) histogram.
           If None is given, bin is computed as :math:`2 \\pi / (Q_{max}-Q_{min})`.
        #. scaleFactor (number): A normalization scale factor used to normalize the computed data to the experimental ones.
        #. adjustScaleFactor (list, tuple): Used to adjust fit or guess the best scale factor during EMC runtime. 
           It must be a list of exactly three entries.\n
           1. The frequency in number of generated moves of finding the best scale factor. 
              If 0 frequency is given, it means that the scale factor is fixed.
           2. The minimum allowed scale factor value.
           3. The maximum allowed scale factor value.
        #. windowFunction (None, numpy.ndarray): The window function to convolute with the computed pair distribution function
           of the system prior to comparing it with the experimental data. In general, the experimental pair
           distribution function G(r) shows artificial wrinkles, among others the main reason is because G(r) is computed
           by applying a sine Fourier transform to the experimental structure factor S(q). Therefore window function is
           used to best imitate the numerical artefacts in the experimental data.
        #. limits (None, tuple, list): The distance limits to compute the histograms.
           If None, the limits will be automatically set the the min and max distance of the experimental data.
           If not None, a tuple of exactly two items where the first is the minimum distance or None 
           and the second is the maximum distance or None.
    
    **NB**: If adjustScaleFactor first item (frequency) is 0, the scale factor will remain 
    untouched and the limits minimum and maximum won't be checked.
    
    .. code-block:: python
    
        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Constraints.StructureFactorConstraints import StructureFactorConstraint
        
        # create engine 
        ENGINE = Engine(pdb='system.pdb')
        
        # create and add constraint
        SFC = StructureFactorConstraint(engine=None, experimentalData="sq.dat", weighting="atomicNumber")
        ENGINE.add_constraints(SFC)
    
    """
    def __init__(self, engine, experimentalData, 
                       dataWeights=None, weighting="atomicNumber", 
                       rmin=None, rmax=None, dr=None, 
                       scaleFactor=1.0, adjustScaleFactor=(0, 0.8, 1.2), 
                       windowFunction=None, limits=None):
        # initialize variables
        self.__limits              = limits
        self.__experimentalQValues = None
        self.__rmin                = None
        self.__rmax                = None
        self.__dr                  = None
        self.__minimumDistance     = None
        self.__maximumDistance     = None
        self.__bin                 = None
        self.__shellCenters        = None
        self.__histogramSize       = None
        self.__shellVolumes        = None
        self.__Gr2SqMatrix         = None
        # initialize constraint
        super(StructureFactorConstraint, self).__init__(engine=engine, experimentalData=experimentalData, dataWeights=dataWeights, scaleFactor=scaleFactor, adjustScaleFactor=adjustScaleFactor)
        # set elements weighting
        self.set_weighting(weighting)
        self.__set_weighting_scheme()
        # set window function
        self.set_window_function(windowFunction)
        # set r parameters
        self.set_rmin(rmin)
        self.set_rmax(rmax)
        self.set_dr(dr)

    def __getstate__(self):
        # make sure that __Gr2SqMatrix is not pickled but saved to the disk as None
        D = {}
        for k, v in self.__dict__.iteritems():
            if k == "_StructureFactorConstraint__Gr2SqMatrix":
                v = None
            D[k] = v
        return D
        
    def __setstate__(self, d):
        # make sure to regenerate G(r) to S(q) matrix at loading time
        self.__dict__ = d
        self.__set_Gr_2_Sq_matrix()      
        
    def __set_Gr_2_Sq_matrix(self):
        if self.__experimentalQValues is None or self.__shellCenters is None:
            self.__Gr2SqMatrix = None
        else:
            Qs = self.__experimentalQValues
            Rs = self.__shellCenters
            dr = self.__shellCenters[1]-self.__shellCenters[0]
            qr = Rs.reshape((-1,1))*(np.ones((len(Rs),1), dtype=FLOAT_TYPE)*Qs)
            sinqr = np.sin(qr)
            sinqr_q = sinqr/Qs
            self.__Gr2SqMatrix = dr*sinqr_q
        
    def __set_used_data_weights(self, minDistIdx=None, maxDistIdx=None):
        # set used dataWeights
        if self.dataWeights is None:
            self._usedDataWeights = None
        else:
            if minDistIdx is None:
                minDistIdx = 0
            if maxDistIdx is None:
                maxDistIdx = self.experimentalData.shape[0]
            self._usedDataWeights  = np.copy(self.dataWeights[minDistIdx:maxDistIdx+1])
            assert np.sum(self._usedDataWeights), LOGGER.error("used points dataWeights are all zero.")
            self._usedDataWeights /= FLOAT_TYPE( np.sum(self._usedDataWeights) )
            self._usedDataWeights *= FLOAT_TYPE( len(self._usedDataWeights) ) 
            
    def __set_weighting_scheme(self):
        if self.engine is not None:
            self.__elementsPairs   = sorted(itertools.combinations_with_replacement(self.engine.elements,2))
            elementsWeights        = dict([(el,float(get_element_property(el,self.__weighting))) for el in self.engine.elements])
            self.__weightingScheme = get_normalized_weighting(numbers=self.engine.numberOfAtomsPerElement, weights=elementsWeights)
            for k, v in self.__weightingScheme.items():
                self.__weightingScheme[k] = FLOAT_TYPE(v)
        else:
            self.__elementsPairs   = None
            self.__weightingScheme = None
            
    def __set_histogram(self):
        if self.__minimumDistance is None or self.__maximumDistance is None or self.__bin is None:
            self.__shellCenters  = None
            self.__histogramSize = None
            self.__shellVolumes  = None
        else:
            # compute edges
            if self.engine is not None and self.rmax is None:
                minHalfBox = np.min( [np.linalg.norm(v)/2. for v in self.engine.basisVectors])
                self.__edges = np.arange(self.__minimumDistance,minHalfBox, self.__bin).astype(FLOAT_TYPE)
            else:
                self.__edges = np.arange(self.__minimumDistance, self.__maximumDistance+self.__bin, self.__bin).astype(FLOAT_TYPE)
            # adjust rmin and rmax
            self.__minimumDistance  = self.__edges[0]
            self.__maximumDistance  = self.__edges[-1]
            # compute shellCenters
            self.__shellCenters = (self.__edges[0:-1]+self.__edges[1:])/FLOAT_TYPE(2.)
            # set histogram size
            self.__histogramSize = INT_TYPE( len(self.__edges)-1 )
            # set shell centers and volumes
            self.__shellVolumes = FLOAT_TYPE(4.0/3.)*PI*((self.__edges[1:])**3 - self.__edges[0:-1]**3)
        # reset constraint
        self.reset_constraint()
        # reset sq matrix
        self.__set_Gr_2_Sq_matrix()
        
    @property
    def rmin(self):
        """ Get the given histogram minimum distance. """
        return self.__rmin
    
    @property
    def rmax(self):
        """ Get the given histogram maximum distance. """
        return self.__rmax
    
    @property
    def dr(self):
        """ Get the given histogram bin size."""
        return self.__dr
    
    @property
    def bin(self):
        """ Get the computed histogram distance bin size."""
        return self.__bin
          
    @property
    def minimumDistance(self):
        """ Get the computed histogram minimum distance. """
        return self.__minimumDistance
          
    @property
    def maximumDistance(self):
        """ Get the computed histogram maximum distance. """
        return self.__maximumDistance
        
    @property
    def qmin(self):
        """ Get the experimental data reciprocal distances minimum. """
        return self.__qmin
          
    @property
    def qmax(self):
        """ Get the experimental data reciprocal distances maximum. """
        return self.__qmax
    
    @property
    def dq(self):
        """ Get the experimental data reciprocal distances bin size. """
        return self.__experimentalQValues[1]-self.__experimentalQValues[0]
        
    @property
    def experimentalQValues(self):
        """ Gets the experimental data used q values. """
        return self.__experimentalQValues
        
    @property
    def histogramSize(self):
        """ Get the histogram size"""
        return self.__histogramSize
        
    @property
    def shellCenters(self):
        """ Get the shells center array"""
        return self.__shellCenters
        
    @property
    def shellVolumes(self):
        """ Get the shells volume array"""
        return self.__shellVolumes
        
    @property
    def experimentalSF(self):
        """ Get the experimental Structure Factor or S(q)"""
        return self.__experimentalSF
        
    @property
    def elementsPairs(self):
        """ Get elements pairs """
        return self.__elementsPairs
        
    @property
    def weightingScheme(self):
        """ Get elements weighting scheme. """
        return self.__weightingScheme
    
    @property
    def windowFunction(self):
        """ Get the window function. """
        return self.__windowFunction
    
    @property
    def limits(self):
        """ The histogram computation limits."""
        return self.__limits
     
    @property
    def Gr2SqMatrix(self):
        """ Get G(r) to S(q) transformation matrix."""
        return self.__Gr2SqMatrix
                
    def listen(self, message, argument=None):
        """   
        Listens to any message sent from the Broadcaster.
        
        :Parameters:
            #. message (object): Any python object to send to constraint's listen method.
            #. argument (object): Any type of argument to pass to the listeners.
        """
        if message in("engine changed", "update molecules indexes"):
            self.__set_weighting_scheme()
            # reset histogram
            if self.engine is not None:
                self.__set_histogram()
        elif message in("update boundary conditions",):
            self.reset_constraint()
            
    def set_rmin(self, rmin):
        """
        Set rmin value.
        
        :parameters:
            #. rmin (None, number): The minimum distance value to compute G(r) histogram.
               If None is given, rmin is computed as :math:`2 \\pi / Q_{max}`.
        """
        if rmin is None:
            minimumDistance = FLOAT_TYPE( 2.*PI/self.__qmax )
        else:
            assert is_number(rmin), LOGGER.error("rmin must be None or a number")
            minimumDistance = FLOAT_TYPE(rmin)
        if self.__maximumDistance is not None:
            assert minimumDistance<self.__maximumDistance, LOGGER.error("rmin must be smaller than rmax %s"%self.__maximumDistance)
        self.__rmin = rmin
        self.__minimumDistance = minimumDistance
        # reset histogram
        self.__set_histogram()

    def set_rmax(self, rmax):
        """
        Set rmax value.
        
        :Parameters:
           #. rmax (None, number): The maximum distance value to compute G(r) histogram.
              If None is given, rmax is computed as :math:`2 \\pi / dQ`.
        """
        if rmax is None:
            dq = self.__experimentalQValues[1]-self.__experimentalQValues[0]
            maximumDistance = FLOAT_TYPE( 2.*PI/dq )
        else:
            assert is_number(rmax), LOGGER.error("rmax must be None or a number")
            maximumDistance = FLOAT_TYPE(rmax)
        if self.__minimumDistance is not None:
            assert maximumDistance>self.__minimumDistance, LOGGER.error("rmax must be bigger than rmin %s"%self.__minimumDistance)
        self.__rmax = rmax
        self.__maximumDistance = maximumDistance
        # reset histogram
        self.__set_histogram()
    
    def set_dr(self, dr):
        """
        Set dr value.
        
        :Parameters:
            #. dr (None, number): The distance bin value to compute G(r) histogram.
               If None is given, bin is computed as :math:`2 \\pi / (Q_{max}-Q_{min})`.
        """
        if dr is None:
            bin  = 2.*PI/self.__qmax
            rbin = round(bin,1)
            if rbin>bin:
                rbin -= 0.1
            bin = FLOAT_TYPE( rbin  )
        else:
            assert is_number(dr), LOGGER.error("dr must be None or a number")
            bin = FLOAT_TYPE(dr)
        self.__dr = dr
        self.__bin = bin
        # reset histogram
        self.__set_histogram()
    
    def set_weighting(self, weighting):
        """
        Sets elements weighting. It must a valid entry of pdbParser atoms database
        
        :Parameters:
            #. weighting (string): The elements weighting.
        """
        assert is_element_property(weighting),LOGGER.error( "weighting is not a valid pdbParser atoms database entry")
        assert weighting != "atomicFormFactor", LOGGER.error("atomicFormFactor weighting is not allowed")
        self.__weighting = weighting
     
    def set_window_function(self, windowFunction):
        """
        Sets the window function.
        
        :Parameters:
             #. windowFunction (None, numpy.ndarray): The window function to convolute with the computed pair distribution function
                of the system prior to comparing it with the experimental data. In general, the experimental pair
                distribution function G(r) shows artificial wrinkles, among others the main reason is because G(r) is computed
                by applying a sine Fourier transform to the experimental structure factor S(q). Therefore window function is
                used to best imitate the numerical artefacts in the experimental data.
        """
        if windowFunction is not None:
            assert isinstance(windowFunction, np.ndarray), LOGGER.error("windowFunction must be a numpy.ndarray")
            assert windowFunction.dtype.type is FLOAT_TYPE, LOGGER.error("windowFunction type must be %s"%FLOAT_TYPE)
            assert len(windowFunction.shape) == 1, LOGGER.error("windowFunction must be of dimension 1")
            assert len(windowFunction) <= self.experimentalData.shape[0], LOGGER.error("windowFunction length must be smaller than experimental data")
            # normalize window function
            windowFunction /= np.sum(windowFunction)
        # check window size
        # set windowFunction
        self.__windowFunction = windowFunction
    
    def set_experimental_data(self, experimentalData):
        """
        Sets the constraint's experimental data.
        
        :Parameters:
            #. experimentalData (numpy.ndarray, string): The experimental data as numpy.ndarray or string path to load data using numpy.loadtxt.
        """
        # get experimental data
        super(StructureFactorConstraint, self).set_experimental_data(experimentalData=experimentalData)
        # set limits
        self.set_limits(self.__limits)
    
    def set_limits(self, limits):
        """
        Set the reciprocal distance limits (qmin, qmax).
        
        :Parameters:
            #. limits (None, tuple, list): The distance limits to compute the histograms and compute with the experimental data.
               If None, the limits will be automatically set the the min and max reciprocal distance recorded in the experimental data.
               If not None, a tuple of minimum reciprocal distance (qmin) or None and maximum reciprocal distance (qmax) or None should be given.    
        """
        if limits is None:
            self.__limits = (None, None)
        else:
            assert isinstance(limits, (list, tuple)), LOGGER.error("limits must be None or a list")
            limits = list(limits)
            assert len(limits) == 2, LOGGER.error("limits list must have exactly two elements")
            if limits[0] is not None:
                assert is_number(limits[0]), LOGGER.error("if not None, the first limits element must be a number")
                limits[0] = FLOAT_TYPE(limits[0])
                assert limits[0]>=0, LOGGER.error("if not None, the first limits element must be bigger or equal to 0")
            if limits[1] is not None:
                assert is_number(limits[1]), LOGGER.error("if not None, the second limits element must be a number")
                limits[1] = FLOAT_TYPE(limits[1])
                assert limits[1]>0, LOGGER.error("if not None, the second limits element must be a positive number")
            if  limits[0] is not None and limits[1] is not None:
                assert limits[0]<limits[1], LOGGER.error("if not None, the first limits element must be smaller than the second limits element")
            self.__limits = (limits[0], limits[1])
        # get minimumDistance and maximumDistance indexes
        if self.__limits[0] is None:
            minDistIdx = 0
        else:
            minDistIdx = (np.abs(self.experimentalData[:,0]-self.__limits[0])).argmin()
        if self.__limits[1] is None:
            maxDistIdx = self.experimentalData.shape[0]-1
        else:
            maxDistIdx =(np.abs(self.experimentalData[:,0]-self.__limits[1])).argmin()
        # set qvalues
        self.__experimentalQValues = self.experimentalData[minDistIdx:maxDistIdx+1,0].astype(FLOAT_TYPE)
        self.__experimentalSF      = self.experimentalData[minDistIdx:maxDistIdx+1,1].astype(FLOAT_TYPE)    
        # set qmin and qmax
        self.__qmin = self.__experimentalQValues[0]
        self.__qmax = self.__experimentalQValues[-1]
        # set used dataWeights
        self.__set_used_data_weights(minDistIdx=minDistIdx, maxDistIdx=maxDistIdx)   
        # reset constraint
        self.reset_constraint()
        # reset sq matrix
        self.__set_Gr_2_Sq_matrix()
        
    def set_data_weights(self, dataWeights):
        """
        Set experimental data points weight.
        
        :Parameters: 
        
        #. dataWeights (None, numpy.ndarray): A weights array of the same number of points of experimentalData used in the constraint's standard error computation.
           Therefore particular fitting emphasis can be put on different data points that might be considered as more or less
           important in order to get a reasonable and plausible modal.\n
           If None is given, all data points are considered of the same importance in the computation of the constraint's standard error.\n
           If numpy.ndarray is given, all weights must be positive and all zeros weighted data points won't contribute to the 
           total constraint's standard error. At least a single weight point is required to be non-zeros and the weights 
           array will be automatically scaled upon setting such as the the sum of all the weights is equal to the number of data points.       
        """
        super(StructureFactorConstraint, self).set_data_weights(dataWeights=dataWeights)
        self.__set_used_data_weights()
        
    def compute_and_set_standard_error(self):
        """ Computes and sets the constraint's standardError."""
        # set standardError
        totalSQ = self.get_constraint_value()["sf_total"]
        self.set_standard_error(self.compute_standard_error(modelData = totalSQ))
 
    def check_experimental_data(self, experimentalData):
        """
        Check whether experimental data is correct.
 
        :Parameters:
            #. experimentalData (object): The experimental data to check.
 
        :Returns:
            #. result (boolean): Whether it is correct or not.
            #. message (str): Checking message that explains whats's wrong with the given data
        """
        if not isinstance(experimentalData, np.ndarray):
            return False, "experimentalData must be a numpy.ndarray"
        if experimentalData.dtype.type is not FLOAT_TYPE:
            return False, "experimentalData type must be %s"%FLOAT_TYPE
        if len(experimentalData.shape) !=2:
            return False, "experimentalData must be of dimension 2"
        if experimentalData.shape[1] !=2:
            return False, "experimentalData must have only 2 columns"
        # check distances order
        inOrder = (np.array(sorted(experimentalData[:,0]), dtype=FLOAT_TYPE)-experimentalData[:,0])<=PRECISION
        if not np.all(inOrder):
            return False, "experimentalData distances are not sorted in order"
        if experimentalData[0][0]<0:
            return False, "experimentalData distances min value is found negative"
        # data format is correct
        return True, ""
        
    def compute_standard_error(self, modelData):
        """ 
        Compute the standard error (StdErr) as squared deviations
        between model computed data and the experimental ones. 
        
        .. math::
            StdErr = \\sum \\limits_{i}^{N} W_{i}(Y(X_{i})-F(X_{i}))^{2}
         
        Where:\n
        :math:`N` is the total number of experimental data points. \n
        :math:`W_{i}` is the data point weight. It becomes equivalent to 1 when dataWeights is set to None. \n
        :math:`Y(X_{i})` is the experimental data point :math:`X_{i}`. \n
        :math:`F(X_{i})` is the computed from the model data  :math:`X_{i}`. \n

        :Parameters:
            #. modelData (numpy.ndarray): The data to compare with the experimental one and compute the standard error.
            
        :Returns:
            #. standardError (number): The calculated standardError of the constraint.
        """
        # compute difference
        diff = self.__experimentalSF-modelData
        # return standard error
        if self._usedDataWeights is None:
            return np.add.reduce((diff)**2)
        else:
            return np.add.reduce(self._usedDataWeights*((diff)**2))
        
    def _get_Sq_from_Gr(self, Gr):
        return np.sum(Gr.reshape((-1,1))*self.__Gr2SqMatrix, axis=0)+1
    
    def __get_total_Sq(self, data):
        """
        This method is created just to speed up the computation of the total Sq upon fitting.
        """
        Gr = np.zeros(self.__histogramSize, dtype=FLOAT_TYPE)
        for pair in self.__elementsPairs:
            # get weighting scheme
            wij = self.__weightingScheme.get(pair[0]+"-"+pair[1], None)
            if wij is None:
                wij = self.__weightingScheme[pair[1]+"-"+pair[0]]
            # get number of atoms per element
            ni = self.engine.numberOfAtomsPerElement[pair[0]]
            nj = self.engine.numberOfAtomsPerElement[pair[1]]
            # get index of element
            idi = self.engine.elements.index(pair[0])
            idj = self.engine.elements.index(pair[1])
            # get Nij
            if idi == idj:
                Nij = ni*(ni-1)/2.0 
                Dij = Nij/self.engine.volume  
                nij = data["intra"][idi,idj,:]+data["inter"][idi,idj,:]
                Gr += wij*nij/Dij      
            else:
                Nij = ni*nj
                Dij = Nij/self.engine.volume
                nij = data["intra"][idi,idj,:]+data["intra"][idj,idi,:] + data["inter"][idi,idj,:]+data["inter"][idj,idi,:]  
                Gr += wij*nij/Dij
        # Devide by shells volume
        Gr /= self.shellVolumes
        # compute total G(r)
        rho0 = (self.engine.numberOfAtoms/self.engine.volume).astype(FLOAT_TYPE)
        Gr   = (FLOAT_TYPE(4.)*PI*self.__shellCenters*rho0)*( Gr-1)
        # Compute S(q) from G(r)
        Sq = self._get_Sq_from_Gr(Gr)
        # Multiply by scale factor
        self._fittedScaleFactor = self.get_adjusted_scale_factor(self.__experimentalSF, Sq, self._usedDataWeights)
        Sq *= self._fittedScaleFactor
        # convolve total with window function
        if self.__windowFunction is not None:
            Sq = np.convolve(Sq, self.__windowFunction, 'same')
        return Sq
    
    def _get_constraint_value(self, data):
        # http://erice2011.docking.org/upload/Other/Billinge_PDF/03-ReadingMaterial/BillingePDF2011.pdf    page 6
        #import time
        #startTime = time.clock()
        output = {}
        for pair in self.__elementsPairs:
            output["sf_intra_%s-%s" % pair] = np.zeros(self.__histogramSize, dtype=FLOAT_TYPE)
            output["sf_inter_%s-%s" % pair] = np.zeros(self.__histogramSize, dtype=FLOAT_TYPE)
            output["sf_total_%s-%s" % pair] = np.zeros(self.__histogramSize, dtype=FLOAT_TYPE)
        gr = np.zeros(self.__histogramSize, dtype=FLOAT_TYPE)
        for pair in self.__elementsPairs:
            # get weighting scheme
            wij = self.__weightingScheme.get(pair[0]+"-"+pair[1], None)
            if wij is None:
                wij = self.__weightingScheme[pair[1]+"-"+pair[0]]
            # get number of atoms per element
            ni = self.engine.numberOfAtomsPerElement[pair[0]]
            nj = self.engine.numberOfAtomsPerElement[pair[1]]
            # get index of element
            idi = self.engine.elements.index(pair[0])
            idj = self.engine.elements.index(pair[1])
            # get Nij
            if idi == idj:
                Nij = ni*(ni-1)/2.0 
                output["sf_intra_%s-%s" % pair] += data["intra"][idi,idj,:] 
                output["sf_inter_%s-%s" % pair] += data["inter"][idi,idj,:]                
            else:
                Nij = ni*nj
                output["sf_intra_%s-%s" % pair] += data["intra"][idi,idj,:] + data["intra"][idj,idi,:]
                output["sf_inter_%s-%s" % pair] += data["inter"][idi,idj,:] + data["inter"][idj,idi,:]
            # compute g(r)
            nij = output["sf_intra_%s-%s" % pair] + output["sf_inter_%s-%s" % pair]
            dij = nij/self.__shellVolumes
            Dij = Nij/self.engine.volume
            gr += wij*dij/Dij
            # calculate intensityFactor
            intensityFactor = (self.engine.volume*wij)/(Nij*self.__shellVolumes)
            # divide by factor
            output["sf_intra_%s-%s" % pair] *= intensityFactor
            output["sf_inter_%s-%s" % pair] *= intensityFactor
            output["sf_total_%s-%s" % pair]  = output["sf_intra_%s-%s" % pair] + output["sf_inter_%s-%s" % pair]
            # Compute S(q) from G(r)
            output["sf_intra_%s-%s" % pair] = self._get_Sq_from_Gr(output["sf_intra_%s-%s" % pair])
            output["sf_inter_%s-%s" % pair] = self._get_Sq_from_Gr(output["sf_inter_%s-%s" % pair])
            output["sf_total_%s-%s" % pair] = self._get_Sq_from_Gr(output["sf_total_%s-%s" % pair])
        # compute total G(r)
        rho0 = (self.engine.numberOfAtoms/self.engine.volume).astype(FLOAT_TYPE)
        Gr = (FLOAT_TYPE(4.)*PI*self.__shellCenters*rho0) * (gr-1)
        # Compute S(q) from G(r)
        Sq = self._get_Sq_from_Gr(Gr)
        output["sf_total"] = self.scaleFactor * Sq
        # convolve total with window function
        if self.__windowFunction is not None:
            output["sf"] = np.convolve(output["sf_total"], self.__windowFunction, 'same').astype(FLOAT_TYPE)
        else:
            output["sf"] = output["sf_total"]
        return output
    
    def get_constraint_value(self):
        """
        Compute all partial Pair Distribution Functions (PDFs). 
        
        :Returns:
            #. PDFs (dictionary): The PDFs dictionnary, where keys are the element wise intra and inter molecular PDFs and values are the computed PDFs.
        """
        if self.data is None:
            LOGGER.warn("data must be computed first using 'compute_data' method.")
            return {}
        return self._get_constraint_value(self.data)
    
    def get_constraint_original_value(self):
        """
        Compute all partial Pair Distribution Functions (PDFs). 
        
        :Returns:
            #. PDFs (dictionary): The PDFs dictionnary, where keys are the element wise intra and inter molecular PDFs and values are the computed PDFs.
        """
        if self.originalData is None:
            LOGGER.warn("originalData must be computed first using 'compute_data' method.")
            return {}
        return self._get_constraint_value(self.originalData)
        
    def compute_data(self):
        """ Compute data and update engine constraintsData dictionary. """
        intra,inter = full_pairs_histograms_coords( boxCoords        = self.engine.boxCoordinates,
                                                    basis            = self.engine.basisVectors,
                                                    isPBC            = self.engine.isPBC,
                                                    moleculeIndex    = self.engine.moleculesIndexes,
                                                    elementIndex     = self.engine.elementsIndexes,
                                                    numberOfElements = self.engine.numberOfElements,
                                                    minDistance      = self.__minimumDistance,
                                                    maxDistance      = self.__maximumDistance,
                                                    histSize         = self.__histogramSize,
                                                    bin              = self.__bin,
                                                    ncores           = self.engine._runtime_ncores  )                  
        # update data
        self.set_data({"intra":intra, "inter":inter})
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # set standardError
        totalPDF = self.__get_total_Sq(self.data)
        self.set_standard_error(self.compute_standard_error(modelData = totalPDF))
    
    def compute_before_move(self, indexes):
        """ 
        Compute constraint before move is executed
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        intraM,interM = multiple_pairs_histograms_coords( indexes          = indexes,
                                                          boxCoords        = self.engine.boxCoordinates,
                                                          basis            = self.engine.basisVectors,
                                                          isPBC            = self.engine.isPBC,
                                                          moleculeIndex    = self.engine.moleculesIndexes,
                                                          elementIndex     = self.engine.elementsIndexes,
                                                          numberOfElements = self.engine.numberOfElements,
                                                          minDistance      = self.__minimumDistance,
                                                          maxDistance      = self.__maximumDistance,
                                                          histSize         = self.__histogramSize,
                                                          bin              = self.__bin,
                                                          allAtoms         = True,
                                                          ncores           = self.engine._runtime_ncores )  
        intraF,interF = full_pairs_histograms_coords( boxCoords        = self.engine.boxCoordinates[indexes],
                                                      basis            = self.engine.basisVectors,
                                                      isPBC            = self.engine.isPBC,
                                                      moleculeIndex    = self.engine.moleculesIndexes[indexes],
                                                      elementIndex     = self.engine.elementsIndexes[indexes],
                                                      numberOfElements = self.engine.numberOfElements,
                                                      minDistance      = self.__minimumDistance,
                                                      maxDistance      = self.__maximumDistance,
                                                      histSize         = self.__histogramSize,
                                                      bin              = self.__bin,
                                                      ncores           = self.engine._runtime_ncores )                                             
        self.set_active_atoms_data_before_move( {"intra":intraM-intraF, "inter":interM-interF} )
        self.set_active_atoms_data_after_move(None)
    
    def compute_after_move(self, indexes, movedBoxCoordinates):
        """ 
        Compute constraint after move is executed
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to.
            #. movedBoxCoordinates (numpy.ndarray): The moved atoms new coordinates.
        """
        # change coordinates temporarily
        boxData = np.array(self.engine.boxCoordinates[indexes], dtype=FLOAT_TYPE)
        self.engine.boxCoordinates[indexes] = movedBoxCoordinates
        # calculate pair distribution function
        intraM,interM = multiple_pairs_histograms_coords( indexes          = indexes,
                                                          boxCoords        = self.engine.boxCoordinates,
                                                          basis            = self.engine.basisVectors,
                                                          isPBC            = self.engine.isPBC,
                                                          moleculeIndex    = self.engine.moleculesIndexes,
                                                          elementIndex     = self.engine.elementsIndexes,
                                                          numberOfElements = self.engine.numberOfElements,
                                                          minDistance      = self.__minimumDistance,
                                                          maxDistance      = self.__maximumDistance,
                                                          histSize         = self.__histogramSize,
                                                          bin              = self.__bin,
                                                          allAtoms         = True,
                                                          ncores           = self.engine._runtime_ncores )  
        intraF,interF = full_pairs_histograms_coords( boxCoords        = self.engine.boxCoordinates[indexes],
                                                      basis            = self.engine.basisVectors,
                                                      isPBC            = self.engine.isPBC,
                                                      moleculeIndex    = self.engine.moleculesIndexes[indexes],
                                                      elementIndex     = self.engine.elementsIndexes[indexes],
                                                      numberOfElements = self.engine.numberOfElements,
                                                      minDistance      = self.__minimumDistance,
                                                      maxDistance      = self.__maximumDistance,
                                                      histSize         = self.__histogramSize,
                                                      bin              = self.__bin,
                                                      ncores           = self.engine._runtime_ncores  )                                             
        # set active atoms data
        self.set_active_atoms_data_after_move( {"intra":intraM-intraF, "inter":interM-interF} )
        # reset coordinates
        self.engine.boxCoordinates[indexes] = boxData
        # compute standardError after move
        dataIntra = self.data["intra"]-self.activeAtomsDataBeforeMove["intra"]+self.activeAtomsDataAfterMove["intra"]
        dataInter = self.data["inter"]-self.activeAtomsDataBeforeMove["inter"]+self.activeAtomsDataAfterMove["inter"]
        totalPDF = self.__get_total_Sq({"intra":dataIntra, "inter":dataInter})
        self.set_after_move_standard_error( self.compute_standard_error(modelData = totalPDF) )
    
    def accept_move(self, indexes):
        """ 
        Accept move
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        dataIntra = self.data["intra"]-self.activeAtomsDataBeforeMove["intra"]+self.activeAtomsDataAfterMove["intra"]
        dataInter = self.data["inter"]-self.activeAtomsDataBeforeMove["inter"]+self.activeAtomsDataAfterMove["inter"]
        # change permanently _data
        self.set_data( {"intra":dataIntra, "inter":dataInter} )
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # update standardError
        self.set_standard_error( self.afterMoveStandardError )
        self.set_after_move_standard_error( None )
        # set new scale factor
        self._set_fitted_scale_factor_value(self._fittedScaleFactor)
    
    def reject_move(self, indexes):
        """ 
        Reject move
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # update standardError
        self.set_after_move_standard_error( None )
   
    def plot(self, ax=None, intra=True, inter=True, 
                   xlabel=True, xlabelSize=16,
                   ylabel=True, ylabelSize=16,
                   legend=True, legendCols=2, legendLoc='best',
                   title=True, titleStdErr=True, titleScaleFactor=True):
        """ 
        Plot structure factor constraint.
        
        :Parameters:
            #. ax (None, matplotlib Axes): matplotlib Axes instance to plot in.
               If ax is given, the figure won't be rendered and drawn.
               If None is given a new plot figure will be created and the figue will be rendered and drawn.
            #. intra (boolean): Whether to add intra-molecular pair distribution function features to the plot.
            #. inter (boolean): Whether to add inter-molecular pair distribution function features to the plot.
            #. xlabel (boolean): Whether to create x label.
            #. xlabelSize (number): The x label font size.
            #. ylabel (boolean): Whether to create y label.
            #. ylabelSize (number): The y label font size.
            #. legend (boolean): Whether to create the legend or not
            #. legendCols (integer): Legend number of columns.
            #. legendLoc (string): The legend location. Anything among
               'right', 'center left', 'upper right', 'lower right', 'best', 'center', 
               'lower left', 'center right', 'upper left', 'upper center', 'lower center'
               is accepted.
            #. title (boolean): Whether to create the title or not
            #. titleStdErr (boolean): Whether to show constraint standard error value in title.
            #. titleScaleFactor (boolean): Whether to show contraint's scale factor value in title.
        
        :Returns:
            #. axes (matplotlib Axes): The matplotlib axes.
        """
        # get constraint value
        output = self.get_constraint_value()
        if not len(output):
            LOGGER.warn("%s constraint data are not computed."%(self.__class__.__name__))
            return
        # import matplotlib
        import matplotlib.pyplot as plt
        # get axes
        if ax is None:
            AXES = plt.gca()
        else:
            AXES = ax   
        # Create plotting styles
        COLORS  = ["b",'g','r','c','y','m']
        MARKERS = ["",'.','+','^','|']
        INTRA_STYLES = [r[0] + r[1]for r in itertools.product(['--'], list(reversed(COLORS)))]
        INTRA_STYLES = [r[0] + r[1]for r in itertools.product(MARKERS, INTRA_STYLES)]
        INTER_STYLES = [r[0] + r[1]for r in itertools.product(['-'], COLORS)]
        INTER_STYLES = [r[0] + r[1]for r in itertools.product(MARKERS, INTER_STYLES)]
        # plot experimental
        AXES.plot(self.__experimentalQValues,self.experimentalSF, 'ro', label="experimental", markersize=7.5, markevery=1 )
        AXES.plot(self.__experimentalQValues, output["sf"], 'k', linewidth=3.0,  markevery=25, label="total" )
        # plot without window function
        if self.windowFunction is not None:
            AXES.plot(self.__experimentalQValues, output["sf_total"], 'k', linewidth=1.0,  markevery=5, label="total - no window" )
        # plot partials
        intraStyleIndex = 0
        interStyleIndex = 0
        for key, val in output.items():
            if key in ("sf_total", "sf"):
                continue
            elif "intra" in key and intra:
                AXES.plot(self.__experimentalQValues, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key )
                intraStyleIndex+=1
            elif "inter" in key and inter:
                AXES.plot(self.__experimentalQValues, val, INTER_STYLES[interStyleIndex], markevery=5, label=key )
                interStyleIndex+=1
        # plot legend
        if legend:
            AXES.legend(frameon=False, ncol=legendCols, loc=legendLoc)
        # set title
        if title:
            t = ''
            if titleStdErr and self.standardError is not None:
                t += "$std$ $error=%.6f$ "%(self.standardError)
            if titleScaleFactor:
                t += " - "*(len(t)>0) + "$scale$ $factor=%.6f$"%(self.scaleFactor)
            if len(t):
                AXES.set_title(t)
        # set axis labels
        if xlabel:
            AXES.set_xlabel("$Q(\AA^{-1})$", size=xlabelSize)
        if ylabel:
            AXES.set_ylabel("$S(Q)$"  , size=ylabelSize)
        # set background color
        plt.gcf().patch.set_facecolor('white')
        #show
        if ax is None:
            plt.show()
        return AXES
        
class ReducedStructureFactorConstraint(StructureFactorConstraint):
    """
    The Reduced Structure Factor that we will also note S(Q) 
    is exactly the same quantity as the Structure Factor but with 
    the slight difference that it is normalized to 0 rather than 1 
    and therefore :math:`<S(Q)>=0`.
    
    The computation of S(Q) is done through a Sine inverse Fourier transform  
    of the computed pair distribution function noted as G(r).
    
    .. math::
        
        S(Q) = \\frac{1}{Q} \\int_{0}^{\\infty} G(r) sin(Qr) dr
        
    The only reason why the Reduced Structure Factor is implemented, is because
    many experimental data are treated in this form. And it is just convenient not
    to manipulate the experimental data every time.
    """
    def _get_Sq_from_Gr(self, Gr):
        return np.sum(Gr.reshape((-1,1))*self.Gr2SqMatrix, axis=0)
        
        


"""
This is a C compiled module to compute atomic bonds.
"""                      
from libc.math cimport sqrt, abs
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
from cython.parallel import prange
from fullrmc.Core.pairs_distances import pairs_distances_to_point

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32

# declare constants
cdef C_FLOAT32 FLOAT_ZERO      = 0.0
cdef C_FLOAT32 FLOAT_ONE       = 1.0
cdef C_INT32   INT32_ZERO      = 0
cdef C_INT32   INT32_ONE       = 1



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef C_FLOAT32 _single_atom_single_shell_dists( C_FLOAT32[:] distances,
                                                C_FLOAT32    lowerShell,
                                                C_FLOAT32    upperShell,
                                                C_INT32      ncores = 1) nogil:
    # declare variables
    cdef C_INT32 i
    cdef C_FLOAT32 coordNumber = FLOAT_ZERO
    cdef C_INT32 num_threads = ncores
    # loop
    for i in prange(INT32_ZERO, <C_INT32>distances.shape[0], INT32_ONE, nogil=True, schedule="static", num_threads=num_threads):
    #for i from 0 <= i < <C_INT32>distances.shape[0]:
        if lowerShell <= distances[i] <= upperShell:      
            coordNumber += FLOAT_ONE
            #with gil:
            #    print i, lowerShell, upperShell, distances[i]
    return coordNumber



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_single_shell_subdists( C_FLOAT32[:] distances,
                                       C_FLOAT32    lowerShell,
                                       C_FLOAT32    upperShell,
                                       C_INT32      ncores = 1):
    return _single_atom_single_shell_dists( distances  = distances,
                                            lowerShell = lowerShell,
                                            upperShell = upperShell,
                                            ncores     = ncores)


    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_single_shell_totdists( ndarray[C_FLOAT32, ndim=1] distances,
                                       C_INT32[:]                 shellIndexes,
                                       C_FLOAT32                  lowerShell,
                                       C_FLOAT32                  upperShell,
                                       C_INT32                    ncores = 1):
    # declare variables
    return _single_atom_single_shell_dists( distances  = distances[shellIndexes],
                                            lowerShell = lowerShell,
                                            upperShell = upperShell,
                                            ncores     =  ncores)
    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_single_shell_coords( C_INT32                    coreIndex,
                                     C_INT32[:]                 shellIndexes,
                                     ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC,
                                     C_FLOAT32                  lowerShell,
                                     C_FLOAT32                  upperShell,
                                     C_INT32                    ncores = 1):
    # declare variables
    distances = pairs_distances_to_point( point  = boxCoords[ coreIndex ], 
                                          coords = boxCoords[ shellIndexes ],
                                          basis  = basis,
                                          isPBC  = isPBC,
                                          ncores = ncores)  
    # compute and return
    return single_atom_single_shell_subdists(distances  = distances,
                                             lowerShell = lowerShell,
                                             upperShell = upperShell,
                                             ncores     = ncores)    
    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_multi_shells_totdists( ndarray[C_FLOAT32, ndim=1] distances,
                                       list                       shellsIndexes,
                                       C_FLOAT32[:]               lowerShells,
                                       C_FLOAT32[:]               upperShells,
                                       C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] coordNumbers = np.zeros(len(shellsIndexes), dtype=NUMPY_FLOAT32)
    # compute coordination numbers
    for i from 0 <= i < <C_INT32>len(shellsIndexes):
        coordNumbers[i] = _single_atom_single_shell_dists( distances   = distances[shellsIndexes[i]],
                                                           lowerShell  = lowerShells[i],
                                                           upperShell  = upperShells[i],
                                                           ncores      = ncores)
    return coordNumbers
    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_multi_shells_coords( C_INT32                    coreIndex,
                                     list                       shellsIndexes,
                                     ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC,
                                     C_FLOAT32[:]               lowerShells,
                                     C_FLOAT32[:]               upperShells,
                                     C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] coordNumbers = np.zeros(len(shellsIndexes), dtype=NUMPY_FLOAT32)
    # compute coordination numbers
    for i from 0 <= i < <C_INT32>len(shellsIndexes):
        # declare variables
        distances = pairs_distances_to_point( point  = boxCoords[ coreIndex ], 
                                              coords = boxCoords[ shellsIndexes[i] ],
                                              basis  = basis,
                                              isPBC  = isPBC,
                                              ncores = ncores)      
        coordNumbers[i] = _single_atom_single_shell_dists( distances   = distances,
                                                           lowerShell  = lowerShells[i],
                                                           upperShell  = upperShells[i],
                                                           ncores      = ncores)
    return coordNumbers        
    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_coord_number_totdists( C_INT32                    atomIndex,
                                       ndarray[C_FLOAT32, ndim=1] distances,
                                       list                       coresIndexes,
                                       list                       shellsIndexes,
                                       list                       lowerShells,
                                       list                       upperShells,
                                       list                       asCoreDefIdxs,
                                       list                       inShellDefIdxs,
                                       C_FLOAT32[:]               coordNumData,
                                       C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 defIdx
    # compute coordination numbers as core
    for defIdx in asCoreDefIdxs[atomIndex]:
        coordNumber = single_atom_single_shell_totdists(distances    = distances,
                                                        shellIndexes = shellsIndexes[defIdx],
                                                        lowerShell   = lowerShells[defIdx] ,
                                                        upperShell   = upperShells[defIdx],
                                                        ncores       = ncores)
        coordNumData[defIdx] += coordNumber
    # compute coordination numbers in shell
    for defIdx in asCoreDefIdxs[atomIndex]:
        coordNumber = single_atom_single_shell_totdists(distances    = distances,
                                                        shellIndexes = coresIndexes[defIdx],
                                                        lowerShell   = lowerShells[defIdx] ,
                                                        upperShell   = upperShells[defIdx],
                                                        ncores       = ncores)
        coordNumData[defIdx] += coordNumber
    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_coord_number_coords( C_INT32                    atomIndex,
                                     ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC,
                                     list                       coresIndexes,
                                     list                       shellsIndexes,
                                     list                       lowerShells,
                                     list                       upperShells,
                                     list                       asCoreDefIdxs,
                                     list                       inShellDefIdxs,
                                     C_FLOAT32[:]               coordNumData,
                                     C_INT32                    ncores = 1):
    # compute coordination numbers as core
    for defIdx in asCoreDefIdxs[atomIndex]:
        coordNumber = single_atom_single_shell_coords( coreIndex    = atomIndex,
                                                       shellIndexes = shellsIndexes[defIdx],
                                                       boxCoords    = boxCoords,
                                                       basis        = basis,
                                                       isPBC        = isPBC,
                                                       lowerShell   = lowerShells[defIdx],
                                                       upperShell   = upperShells[defIdx],
                                                       ncores       = ncores)
        coordNumData[defIdx] += coordNumber
    # compute coordination numbers in shell
    for defIdx in inShellDefIdxs[atomIndex]:
        coordNumber = single_atom_single_shell_coords( coreIndex    = atomIndex,
                                                       shellIndexes = coresIndexes[defIdx],
                                                       boxCoords    = boxCoords,
                                                       basis        = basis,
                                                       isPBC        = isPBC,
                                                       lowerShell   = lowerShells[defIdx],
                                                       upperShell   = upperShells[defIdx],
                                                       ncores       = ncores)
        coordNumData[defIdx] += coordNumber    
        
        
            
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def multi_atoms_coord_number_totdists( C_INT32[:]   indexes,
                                       list         distances,
                                       list         coresIndexes,
                                       list         shellsIndexes,
                                       list         lowerShells,
                                       list         upperShells,
                                       list         asCoreDefIdxs,
                                       list         inShellDefIdxs,
                                       C_FLOAT32[:] coordNumData,
                                       C_INT32      ncores = 1):
    # declare variables
    cdef C_INT32 i
    for i from 0 <= i < <C_INT32>len(indexes):
        single_atom_coord_number_totdists( atomIndex      = indexes[i],
                                           distances      = distances[i],
                                           coresIndexes   = coresIndexes,
                                           shellsIndexes  = shellsIndexes,
                                           lowerShells    = lowerShells,
                                           upperShells    = upperShells,
                                           asCoreDefIdxs  = asCoreDefIdxs,
                                           inShellDefIdxs = inShellDefIdxs,
                                           coordNumData   = coordNumData,
                                           ncores         = ncores)
    
    

@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def multi_atoms_coord_number_coords( C_INT32[:]                 indexes,
                                     ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC,
                                     list                       coresIndexes,
                                     list                       shellsIndexes,
                                     list                       lowerShells,
                                     list                       upperShells,
                                     list                       asCoreDefIdxs,
                                     list                       inShellDefIdxs,
                                     C_FLOAT32[:]               coordNumData,
                                     C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    for i from 0 <= i < <C_INT32>len(indexes):
        single_atom_coord_number_coords( atomIndex      = indexes[i],
                                         boxCoords      = boxCoords,
                                         basis          = basis,
                                         isPBC          = isPBC,
                                         coresIndexes   = coresIndexes,
                                         shellsIndexes  = shellsIndexes,
                                         lowerShells    = lowerShells,
                                         upperShells    = upperShells,
                                         asCoreDefIdxs  = asCoreDefIdxs,
                                         inShellDefIdxs = inShellDefIdxs,
                                         coordNumData   = coordNumData,
                                         ncores         = ncores)
    
    

    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def all_atoms_coord_number_totdists( ndarray[C_FLOAT32, ndim=2] distances,
                                     list                       coresIndexes,
                                     list                       shellsIndexes,
                                     list                       lowerShells,
                                     list                       upperShells,
                                     list                       asCoreDefIdxs,
                                     list                       inShellDefIdxs,
                                     C_FLOAT32[:]               coordNumData,
                                     C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    # get indexes
    cdef ndarray[C_INT32,  mode="c", ndim=1] indexes = np.arange(<C_INT32>distances.shape[1], dtype=NUMPY_INT32)
    # run multiple atoms coordination number using distances
    multi_atoms_coord_number_totdists( indexes        = indexes,
                                       distances      = distances,
                                       coresIndexes   = coresIndexes,
                                       shellsIndexes  = shellsIndexes,
                                       lowerShells    = lowerShells,
                                       upperShells    = upperShells,
                                       asCoreDefIdxs  = asCoreDefIdxs,
                                       inShellDefIdxs = inShellDefIdxs,
                                       coordNumData   = coordNumData,
                                       ncores         = ncores)
    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def all_atoms_coord_number_coords( ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                   C_FLOAT32[:,:]             basis not None,
                                   bint                       isPBC,
                                   list                       coresIndexes,
                                   list                       shellsIndexes,
                                   list                       lowerShells,
                                   list                       upperShells,
                                   list                       asCoreDefIdxs,
                                   list                       inShellDefIdxs,
                                   C_FLOAT32[:]               coordNumData,
                                   C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    # get indexes
    cdef ndarray[C_INT32,  mode="c", ndim=1] indexes = np.arange(<C_INT32>boxCoords.shape[0], dtype=NUMPY_INT32)
    # run multiple atoms coordination number using coordinates
    multi_atoms_coord_number_coords( indexes        = indexes,
                                     boxCoords      = boxCoords,
                                     basis          = basis,
                                     isPBC          = isPBC,
                                     coresIndexes   = coresIndexes,
                                     shellsIndexes  = shellsIndexes,
                                     lowerShells    = lowerShells,
                                     upperShells    = upperShells,
                                     asCoreDefIdxs  = asCoreDefIdxs,
                                     inShellDefIdxs = inShellDefIdxs,
                                     coordNumData   = coordNumData,
                                     ncores         = ncores)


            
    
    
    
    


"""
Agitations contains all MoveGenerator classes that agitate and shake structures such as distances, angles, etc.

.. inheritance-diagram:: fullrmc.Generators.Agitations
    :parts: 1 

+------------------------------------------------------+------------------------------------------------------+
|.. figure:: distanceAgitation.png                     |.. figure:: angleAgitation.png                        |
|   :width: 375px                                      |   :width: 375px                                      |
|   :height: 300px                                     |   :height: 300px                                     |
|   :align: left                                       |   :align: left                                       |
|                                                      |                                                      |
|   Random H-H bond length agitations generated on     |   Random H-O-H angle agitation generated on water    |
|   dihydrogen molecules. At room temperature, H2      |   molecules. At room temperature, water molecule     |
|   molecule bond length fluctuates around 0.74        |   angle formed between the two vectors formed between|
|   Angstroms. Red hydrogen atoms represent the shrank |   consecutively the Oxygen atom and the two hydrogen |
|   H-H bond length molecule while blue hydrogen atoms |   atoms is about 105 deg. Shrank H-O-H angles are    |
|   represent the expanded H-H bond length molecules.  |   represented by the red hydrogen while expanded     |
|                                                      |   angles are represented in blue.                    |
|   (:class:`DistanceAgitationGenerator`)              |   (:class:`AngleAgitationGenerator`)                 |
+------------------------------------------------------+------------------------------------------------------+
 
 .. raw:: html

        <iframe width="560" height="315" 
        src="https://www.youtube.com/embed/qTJux9kZCOo?rel=0" 
        frameborder="0" allowfullscreen>
        </iframe>
        
"""

# standard libraries imports

# external libraries imports
import numpy as np

# fullrmc imports
from fullrmc.Globals import INT_TYPE, FLOAT_TYPE, PI, LOGGER
from fullrmc.Core.Collection import is_number, is_integer, get_rotation_matrix, generate_random_float
from fullrmc.Core.MoveGenerator import MoveGenerator, PathGenerator


class DistanceAgitationGenerator(MoveGenerator):
    """
    Generates random agitation moves upon a distance separating two atoms by translating
    both atoms away from each other or closer to each other along the direction line between them.
    This is mainly used to shake two atoms bond distance by increasing and decreasing 
    the bond length.
    Only groups of length 2 are accepted.
     
    :Parameters:
        #. group (None, Group): The group instance. It must contain exactly 2 indexes.
        #. amplitude (number):  The maximum translation amplitude in Angstroms applied on every atom.
        #. symmetric (bool): Whether to apply the same amplitude of translation on both atoms or not.
        #. shrink (None, bool): Whether to always shrink the distance or expand it.
           If True, moves will always bring atoms closer to each other.
           If False, moves will always bring atoms away from each other.
           If None, no orientation is forced, therefore atoms can randomly get closer to each other or away from each other.     
        #. agitate (tuple): It's a tuple of two boolean values, at least one of them must be True.
           Whether to agitate the first atom, the second or both. This is useful to set an atom fixed while only 
           the other succumb the agitation to adjust the distance. For instance in a C-H group it can be useful and 
           logical to adjust the bond length by moving only the hydrogen atom along the bond direction.
                   
    .. code-block:: python
    
        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Agitations import DistanceAgitationGenerator
        
        # create engine 
        ENGINE = Engine(pdb='system.pdb')
        
        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...
        
        # set moves generators to random agitations of distance seperating two atoms.
        # Maximum agitation amplitude is set to to 0.5A
        for g in ENGINE.groups:
            if len(g)==2:
                g.set_move_generator( DistanceAgitationGenerator(amplitude=0.5) )
                
    """
    def __init__(self, group=None, amplitude=0.2, symmetric=True, shrink=None, agitate=(True,True)):
        super(DistanceAgitationGenerator, self).__init__(group=group)
        # set amplitude
        self.set_amplitude(amplitude)
        # set symmetric
        self.set_symmetric(symmetric)
        # set shrink
        self.set_shrink(shrink)
        # set agitated
        self.set_agitate(agitate)
        
    @property
    def amplitude(self):
        """Gets the maximum agitation amplitude."""
        return self.__amplitude
    
    @property
    def shrink(self):
        """Gets shrink flag value."""
        return self.__shrink
    
    @property
    def symmetric(self):
        """Gets symmetric flag value."""
        return self.__symmetric
    
    @property
    def agitate(self):
        """Gets agitate tuple flags value."""
        return self.__agitate
        
    def check_group(self, group):
        """
        Checks the generator's group.
        
        :Parameters:
            #. group (Group): the Group instance.
        """
        if len(group.indexes)!=2:
            return False, "two atoms are needed in a group to perform distance agitation movements."
        else:
            return True, "" 
            
    def set_amplitude(self, amplitude):
        """
        Sets maximum translation vector allowed amplitude.
        
        :Parameters:
            #. amplitude (number): the maximum allowed translation vector amplitude.
        """
        assert is_number(amplitude), LOGGER.error("Translation amplitude must be a number")
        amplitude = float(amplitude)
        assert amplitude>0, LOGGER.error("Translation amplitude must be bigger than 0")
        self.__amplitude = FLOAT_TYPE(amplitude)
    
    def set_symmetric(self, symmetric):
        """
        Sets symmetric flag value.
        
        :Parameters:
            #. symmetric (bool): Whether to apply the same amplitude of translation on both atoms or not.         
        """
        assert isinstance(symmetric, bool), LOGGER.error("symmetric must be boolean")
        self.__symmetric = symmetric
    
    def set_shrink(self, shrink):
        """
        Sets shrink flag value.
        
        :Parameters:
            #. shrink (None, bool): Whether to always shrink the distance or expand it.
               If True, moves will always bring atoms closer to each other.
               If False, moves will always bring atoms away from each other.
               If None, no orientation is forced, therefore distance can increase or decrease randomly at every step.           
        """
        assert shrink in (None, True, False), LOGGER.error("shrink can only be None, True or False")
        self.__shrink = shrink
    
    def set_agitate(self, agitate):
        """
        Sets agitate tuple value.
        
        :Parameters:
            #. agitate (tuple): It's a tuple of two boolean values, at least one of them must be True.
               Whether to agitate the first atom, the second or both. This is useful to set an atom fixed while only 
               the other succumb the agitation to adjust the distance. For instance in a C-H group it can be useful and 
               logical to adjust the bond length by moving only the hydrogen atom along the bond direction.
        """
        assert isinstance(agitate, (list,tuple)), LOGGER.error("agitate must be a list or a tuple")
        assert len(agitate)==2, LOGGER.error("agitate must have 2 items")
        assert [isinstance(a,bool) for a in agitate]==[True,True], LOGGER.error("agitate items must be boolean")
        assert agitate[0] or agitate[1], LOGGER.error("agitate both items can't be False")
        self.__agitate = (agitate[0], agitate[1])     

    def transform_coordinates(self, coordinates, argument=None):
        """
        Translate coordinates.
        
        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply the translation.
            
        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying the translation.
            #. argument (object): Any python object. Not used in this generator.
        """
        # get normalized direction vector
        vector  = FLOAT_TYPE( coordinates[0,:]-coordinates[1,:] )
        vector /= FLOAT_TYPE( np.linalg.norm(vector) )
        # create amplitudes
        if self.__symmetric:
            amp0 = amp1 =  FLOAT_TYPE(generate_random_float()*self.__amplitude)
        else:
            amp0 =  FLOAT_TYPE(generate_random_float()*self.__amplitude)
            amp1 =  FLOAT_TYPE(generate_random_float()*self.__amplitude)
        # create shrink flag
        if self.__shrink is None:
            shrink = (1-2*generate_random_float())>0
        else:
            shrink = self.__shrink    
        # create directions
        if shrink:
            dir0 = FLOAT_TYPE(-1)
            dir1 = FLOAT_TYPE( 1)            
        else:
            dir0 = FLOAT_TYPE( 1)
            dir1 = FLOAT_TYPE(-1) 
        # create translation vectors 
        translationVectors      = np.empty((2,3), dtype=FLOAT_TYPE)
        translationVectors[0,:] = self.__agitate[0]*dir0*amp0*vector
        translationVectors[1,:] = self.__agitate[1]*dir1*amp1*vector
        # translate and return
        return coordinates+translationVectors
 

class AngleAgitationGenerator(MoveGenerator):
    """
    Generates random agitation moves upon an angle defined between two vectors left-central and right-central 
    where (central, left, right) are three atoms. Move will be performed on left and/or right atom while 
    central atom will always remain fixed. Distances between left/right and central atoms will remain
    unchanged. This is mainly used to shake bonded atoms angles by increasing and decreasing 
    the bond length.
    Only groups of length 3 are accepted.
    
    :Parameters:
        #. group (None, Group): The group instance. It must contain exactly three indexes in respective
           order (central, left, right) atoms indexes.
        #. amplitude (number):  The maximum agitation angle amplitude in degrees of left and right atoms separately.
        #. symmetric (bool): Whether to apply the same amplitude of rotation on both left and right atoms or not.
        #. shrink (None, bool): Whether to always shrink the angle or expand it.
           If True, moves will always reduce angle.
           If False, moves will always increase angle.
           If None, no orientation is forced, therefore angle can randomly get wider or tighter.     
        #. agitate (tuple): It's a tuple of two boolean values for respectively (left, right) atoms, 
           at least one of them must be True. Whether to agitate the left atom, the right or both. 
           This is useful to set an atom fixed while only the other succumb the agitation to adjust the angle.
    
    .. code-block:: python
    
        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Agitations import AngleAgitationGenerator
        
        # create engine 
        ENGINE = Engine(pdb='system.pdb')
        
        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...
        
        # set moves generators to random agitations of the angle formed between
        # one central atom and other two. Maximum agitation amplitude is set to to 10.
        for g in ENGINE.groups:
            if len(g)==3:
                g.set_move_generator( AngleAgitationGenerator(amplitude=10) )
                
    """
    def __init__(self, group=None, amplitude=2, symmetric=True, shrink=None, agitate=(True,True)):
        super(AngleAgitationGenerator, self).__init__(group=group)
        # set amplitude
        self.set_amplitude(amplitude)
        # set symmetric
        self.set_symmetric(symmetric)
        # set shrink
        self.set_shrink(shrink)
        # set agitated
        self.set_agitate(agitate)
        
    @property
    def amplitude(self):
        """Gets the maximum agitation angle amplitude in rad."""
        return self.__amplitude
    
    @property
    def shrink(self):
        """Gets shrink flag value."""
        return self.__shrink
    
    @property
    def symmetric(self):
        """Gets symmetric flag value."""
        return self.__symmetric
    
    @property
    def agitate(self):
        """Gets agitate tuple flags value."""
        return self.__agitate
        
    def check_group(self, group):
        """
        Checks the generator's group.
        
        :Parameters:
            #. group (Group): the Group instance.
        """
        if len(group.indexes)!=3:
            return False, "Exactly three atoms are needed in a group to perform angle agitation movements."
        else:
            return True, "" 
            
    def set_amplitude(self, amplitude):
        """
        Sets maximum allowed agitation rotation angle amplitude in degrees of left and right atoms separately and transforms it to rad.
        
        :Parameters:
            #. amplitude (number):  The maximum agitation angle amplitude in degrees of left and right atoms separately.
        """
        assert is_number(amplitude), LOGGER.error("Agitation angle amplitude must be a number")
        amplitude = float(amplitude)
        assert amplitude>0, LOGGER.error("Agitation angle amplitude must be bigger than 0")
        assert amplitude<=90, LOGGER.error("Agitation angle amplitude must be smaller than 90")
        self.__amplitude = FLOAT_TYPE(amplitude*PI/180.)
    
    def set_symmetric(self, symmetric):
        """
        Sets symmetric flag value.
        
        :Parameters:
            #. symmetric (bool): Whether to apply the same amplitude of translation on both atoms or not.         
        """
        assert isinstance(symmetric, bool), LOGGER.error("symmetric must be boolean")
        self.__symmetric = symmetric
    
    def set_shrink(self, shrink):
        """
        Sets shrink flag value.
        
        :Parameters:
            #. shrink (None, bool): Whether to always shrink the distance or expand it.
               If True, moves will always bring atoms closer to each other.
               If False, moves will always bring atoms away from each other.
               If None, no orientation is forced, therefore distance can increase or decrease randomly at every step.           
        """
        assert shrink in (None, True, False), LOGGER.error("shrink can only be None, True or False")
        self.__shrink = shrink
    
    def set_agitate(self, agitate):
        """
        Sets agitate tuple value.
        
        :Parameters:
            #. agitate (tuple): It's a tuple of two boolean values, at least one of them must be True.
               Whether to agitate the first atom, the second or both. This is useful to set an atom fixed while only 
               the other succumb the agitation to adjust the distance. For instance in a C-H group it can be useful and 
               logical to adjust the bond length by moving only the hydrogen atom along the bond direction.
        """
        assert isinstance(agitate, (list,tuple)), LOGGER.error("agitate must be a list or a tuple")
        assert len(agitate)==2, LOGGER.error("agitate must have 2 items")
        assert [isinstance(a,bool) for a in agitate]==[True,True], LOGGER.error("agitate items must be boolean")
        assert agitate[0] or agitate[1], LOGGER.error("agitate both items can't be False")
        self.__agitate = (agitate[0], agitate[1])     

    def transform_coordinates(self, coordinates, argument=None):
        """
        Translate coordinates.
        
        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply the translation.
            
        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying the translation.
            #. argument (object): Any python object. Not used in this generator.
        """
        # get atoms group center
        center = np.sum(coordinates, 0)/coordinates.shape[0]
        # translate to origin
        rotatedCoordinates = coordinates-center
        # get normalized direction vectors
        leftVector   = FLOAT_TYPE( rotatedCoordinates[1,:]-rotatedCoordinates[0,:] )
        leftVector  /= FLOAT_TYPE( np.linalg.norm(leftVector) )
        rightVector  = FLOAT_TYPE( rotatedCoordinates[2,:]-rotatedCoordinates[0,:] )
        rightVector /= FLOAT_TYPE( np.linalg.norm(rightVector) )
        # get rotation axis
        rotationAxis = np.cross(leftVector, rightVector)
        if rotationAxis[0]==rotationAxis[1]==rotationAxis[2]==0.:
            rotationAxis = np.array(1-2*np.random.random(3), dtype=FLOAT_TYPE)
            rotationAxis /= FLOAT_TYPE( np.linalg.norm(rotationAxis) )
        # create shrink flag
        if self.__shrink is None:
            shrink = (1-2*generate_random_float())>0
        else:
            shrink = self.__shrink    
        # get rotation angles
        if self.__symmetric:
            angleLeft  = angleRight = FLOAT_TYPE(generate_random_float()*self.__amplitude)
        else:
            angleLeft  = FLOAT_TYPE(generate_random_float()*self.__amplitude)
            angleRight = FLOAT_TYPE(generate_random_float()*self.__amplitude)
        # create directions
        if shrink:
            angleLeft  *= FLOAT_TYPE(-1)
            angleRight *= FLOAT_TYPE( 1)            
        else:
            angleLeft  *= FLOAT_TYPE( 1)
            angleRight *= FLOAT_TYPE(-1) 
        # rotate
        if self.__agitate[0]:
            rotationMatrix = get_rotation_matrix(rotationAxis, angleLeft)
            rotatedCoordinates[1,:] = np.dot( rotationMatrix, rotatedCoordinates[1,:])
        if self.__agitate[1]:
            rotationMatrix = get_rotation_matrix(rotationAxis, angleRight)
            rotatedCoordinates[2,:] = np.dot( rotationMatrix, rotatedCoordinates[2,:])
        # translate back from center and return
        return np.array(rotatedCoordinates+center, dtype=FLOAT_TYPE)
        
        
        
        


"""
MoveGenerator contains parent classes for all move generators.
A MoveGenerator sub-class is used at the Engine runtime to generate moves upon selected groups.
Every group has its own MoveGenerator class and definitions, therefore it has become possible
to fully customize how a group of atoms should move.

.. inheritance-diagram:: fullrmc.Core.MoveGenerator
    :parts: 1
"""

# standard libraries imports
import inspect
from random import randint, shuffle

# external libraries imports
import numpy as np

# fullrmc imports
from fullrmc.Globals import INT_TYPE, FLOAT_TYPE, LOGGER
from fullrmc.Core.Collection import is_number, is_integer, get_path, generate_random_float


class MoveGenerator(object):
    """ 
    It is the parent class for all moves generators.
    This class can't be instantiated but its sub-classes might be.
    
    :Parameters:
        #. group (None, Group): The group instance.
    """
    def __init__(self, group=None):
        self.set_group(group)
    
    @property
    def group(self):
        """ Get the group instance."""
        return self.__group
    
    def listen(self, message, argument=None):
        """   
        Listens to any message sent from the Broadcaster.
        
        :Parameters:
            #. message (object): Any python object to send to constraint's listen method.
            #. arguments (object): Any python object.
        """
        pass
        
    def set_group(self, group):
        """
        Set the MoveGenerator group.
        
        :Parameters:
            #. group (None, Group): group instance. 
        """
        if group is not None:
            from fullrmc.Core.Group import Group
            assert isinstance(group, Group), LOGGER.error("group must be a fullrmc Group instance")
            valid, message = self.check_group(group)
            if not valid:
                raise Exception( LOGGER.error("%s"%message) )
        self.__group = group
    
    def check_group(self, group):
        """
        Checks the generator's group. 
        This method must be overloaded in all MoveGenerator sub-classes.
        
        :Parameters:
            #. group (Group): the Group instance
        """
        raise Exception(LOGGER.error("MovesGenerator '%s' method must be overloaded"%inspect.stack()[0][3]))
        
    def transform_coordinates(self, coordinates, argument=None):
        """
        Transform coordinates. This method is called in every move.
        This method must be overloaded in all MoveGenerator sub-classes.
        
        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply the move.
            #. argument (object): Any other argument needed to perform the move.
               In General it's not needed.
            
        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying the move.
        """
        raise Exception(LOGGER.error("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))
        
    def move(self, coordinates):
        """
        Moves coordinates. 
        This method must NOT be overloaded in MoveGenerator sub-classes.
        
        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply the transformation.
            
        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying the transformation.
        """
        return self.transform_coordinates(coordinates=coordinates)


class SwapGenerator(MoveGenerator):
    """ 
    It is a particular move generator that instead of generating a 
    move upon a group of atoms, it will exchange the group atom positions
    with other atoms from swapList. 
    Because the swapList can be big, swapGenerator can be assigned to
    multiple groups at the same time under the condition of all groups
    having the same length.\n
    
    During engine runtime, whenever a swap generator is encountered,
    all sophisticated selection recurrence modes such as (refining, exploring) 
    will be reduced to simple recurrence.\n

    This class can't be instantiated but its sub-classes might be.
    
    :Parameters:
        #. group (None, Group): The group instance.
        #. swapLength (Integer): The swap length that defines the length of the group 
           and the length of the every swap sub-list in swapList.
        #. swapList (None, List): The list of atoms.\n
           If None is given, no swapping or exchanging will be performed.\n
           If List is given, it must contain lists of atom indexes where every 
           sub-list must have the same number of atoms as the group.
    """
    def __init__(self, group=None, swapLength=1, swapList=None):
        super(SwapGenerator, self).__init__(group=group) 
        # set swap length
        self.set_swap_length(swapLength)
        # set swap list
        self.set_swap_list(swapList)
        #  initialize swapping variables
        self.__groupAtomsIndexes = None
        self.__swapAtomsIndexes  = None        
    
    @property
    def swapLength(self):
        """ Get swap length."""
        return self.__swapLength 
    
    @property
    def swapList(self):
        """ Get swap list."""
        return self.__swapList
        
    @property
    def groupAtomsIndexes (self):
        """ Get last selected group atoms indexes."""
        return self.__groupAtomsIndexes 
    
    @property
    def swapAtomsIndexes(self):
        """ Get last swap atoms indexes."""
        return self.__swapAtomsIndexes
        
    def set_swap_length(self, swapLength):
        """
        Set swap length. it will reset swaplist automatically.
    
        :Parameters:
            #. swapLength (Integer): The swap length that defines the length of the group 
               and the length of the every swap sub-list in swapList.
        """   
        assert is_integer(swapLength), LOGGER.error("swapLength must be an integer")
        swapLength = INT_TYPE(swapLength)
        assert swapLength>0, LOGGER.error("swapLength must be bigger than 0")
        self.__swapLength = swapLength
        self.__swapList   = ()
        
    def set_group(self, group):
        """
        Set the MoveGenerator group.
        
        :Parameters:
            #. group (None, Group): group instance. 
        """
        MoveGenerator.set_group(self, group)
        if self.group is not None:
            assert len(self.group) == self.__swapLength, LOGGER.error("SwapGenerator groups length must be equal to swapLength.")

    def set_swap_list(self, swapList):
        """
        Set the swap-list to exchange atom positions from.
        
        :Parameters: 
            #. swapList (None, List): The list of atoms.\n 
               If None is given, no swapping or exchanging will be performed.\n
               If List is given, it must contain lists of atom indexes where every 
               sub-list length must be equal to swapLength.
        """
        if swapList is None:
            self.__swapList = ()
        else:
            SL = []
            assert isinstance(swapList, (list,tuple)), LOGGER.error("swapList must be a list")
            for sl in swapList:
                assert isinstance(sl, (list,tuple)), LOGGER.error("swapList items must be a list")
                subSL = []
                for num in sl:
                    assert is_integer(num), LOGGER.error("swapList sub-list items must be integers")
                    num = INT_TYPE(num)
                    assert num>=0, LOGGER.error("swapList sub-list items must be positive")
                    subSL.append(num)
                assert len(set(subSL))==len(subSL), LOGGER.error("swapList items must not have any redundancy")
                assert len(subSL) == self.__swapLength, LOGGER.error("swapList item length must be equal to swapLength")
                SL.append(np.array(subSL, dtype=INT_TYPE))
            self.__swapList = tuple(SL)
    
    def append_to_swap_list(self, subList):
        """
        append a sub list to swap list
        
        :Parameters: 
            #. subList (List): The sub-list of atom indexes to append to swapList.
        """
        assert isinstance(subList, (list,tuple)), LOGGER.error("subList must be a list")
        subSL = []
        for num in subList:
            assert is_integer(num), LOGGER.error("subList items must be integers")
            num = INT_TYPE(num)
            assert num>=0, LOGGER.error("subList items must be positive")
            subSL.append(num)
        assert len(set(subSL))==len(subSL), LOGGER.error("swapList items must not have any redundancy")
        assert len(subSL) == self.__swapLength, LOGGER.error("swapList item length must be equal to swapLength")
        # append
        self.__swapList = list(self.__swapList)
        self.__swapList.append(subSL)
        self.__swapList = tuple(self.__swapList)
    
    def get_ready_for_move(self, groupAtomsIndexes):  
        """
        Set the swap generator ready to perform a move. Unlike a normal move generator,
        swap generators will affect not only the selected atoms but other atoms as well.
        Therefore at engine runtime, selected atoms will be extended to all affected 
        atoms by the swap.\n
        This method is called automatically upon engine runtime
        to ensure that all affect atoms with the swap are updated.
        
        :Parameters: 
            #. groupAtomsIndexes (numpy.ndarray): The atoms indexes to swap.
        
        :Returns: 
            #. indexes (numpy.ndarray): All the atoms involved in the swap move 
               including the given groupAtomsIndexes.
        """
        self.__groupAtomsIndexes = groupAtomsIndexes
        self.__swapAtomsIndexes  = self.swapList[ randint(0,len(self.swapList)-1) ]
        return np.concatenate( (self.__groupAtomsIndexes,self.__swapAtomsIndexes) )
        
        
        
class PathGenerator(MoveGenerator):
    """ 
    PathGenerator is a MoveGenerator sub-class where moves definitions are pre-stored in a path 
    and get pulled out at every move step.
    This class can't be instantiated but its sub-classes might be.
    
    :Parameters:
        #. group (None, Group): The group instance.
        #. path (None, list): The list of moves.
        #. randomize (boolean): Whether to pull moves randomly from path or pull moves in order at every step.
    """
    
    def __init__(self, group=None, path=None, randomize=False):
        super(PathGenerator, self).__init__(group=group) 
        # set path
        self.set_path(path)
        # set randomize
        self.set_randomize(randomize)
        # initialize flags
        self.__initialize_path_generator__()
    
    def __initialize_path_generator__(self):
        self.__step = 0
        
    @property
    def step(self):
        """ Get the current step number."""
        return self.__step
        
    @property
    def path(self):
        """ Get the path list of moves."""
        return self.__path
        
    @property
    def randomize(self):
        """ Get randomize flag."""
        return self.__randomize
    
    def check_path(self, path):
        """
        Checks the generator's path.
        This method must be overloaded in all PathGenerator sub-classes.
        
        :Parameters:
            #. path (list): The list of moves.
        """ 
        raise Exception(LOGGER.error("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))
    
    def normalize_path(self, path):
        """
        Normalizes all path moves. 
        It is called automatically upon set_path method is called.
        This method can be overloaded in all MoveGenerator sub-classes.
        
        :Parameters:
            #. path (list): The list of moves.
        
        :Returns:
            #. path (list): The list of moves.
        """
        return list(path)
        
    def set_path(self, path):
        """
        Sets the moves path.
        
        :Parameters:
            #. path (list): The list of moves.
        """
        valid, message = self.check_path(path)
        if not valid:
            raise Exception(message)
        # normalize path
        self.__path = self.normalize_path( path )
        # reset generator
        self.__initialize_path_generator__()
        
    def set_randomize(self, randomize):
        """
        Sets whether to randomize moves selection.
        
        :Parameters:
            #. randomize (boolean): Whether to pull moves randomly from path or pull moves in order at every step.
        """
        assert isinstance(randomize, bool), LOGGER.error("randomize must be boolean")
        self.__randomize = randomize
        
    def move(self, coordinates):
        """
        Moves coordinates.
        
        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply the transformation
            
        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying the transformation
        """
        if self.__randomize:
            move = self.__path[ randint(0,len(self.__path)-1) ]
        else:
            move = self.__path[self.__step]
            self.__step = (self.__step+1)%len(self.__path)
        # perform the move
        return self.transform_coordinates(coordinates, argument=move)    

    
class MoveGeneratorCombinator(MoveGenerator):
    """ 
    MoveGeneratorCombinator combines all moves of a list of MoveGenerators and applies it at once.
    
    :Parameters:
        #. group (None, Group): The constraint RMC engine.
        #. combination (list): The list of MoveGenerator instances.
        #. shuffle (boolean): Whether to shuffle generator instances at every move or to combine moves in the list order.
    
    
    .. code-block:: python
    
        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Core.MoveGenerator import MoveGeneratorCombinator
        from fullrmc.Generators.Translations import TranslationGenerator
        from fullrmc.Generators.Rotations import RotationGenerator
        
        # create engine 
        ENGINE = Engine(pdb='system.pdb')
        
        # Add constraints ...
        # Re-define groups if needed ...
        
        ##### Define each group move generator as a combination of a translation and a rotation. #####
        # create recursive group selector. Recurrence is set to 20 with explore flag set to True.
        # shuffle is set to True which means that at every selection the order of move generation
        # is random. At one step a translation is performed prior to rotation and in another step
        # the rotation is performed at first.
        # selected from the collector.
        for g in ENGINE.groups:
            # create translation generator
            TMG = TranslationGenerator(amplitude=0.2)
            # create rotation generator only when group length is bigger than 1.
            if len(g)>1:
                RMG = RotationGenerator(amplitude=2)
                MG  = MoveGeneratorCombinator(collection=[TMG,RMG],shuffle=True)
            else:
                MG  = MoveGeneratorCombinator(collection=[TMG],shuffle=True)
            g.set_move_generator( MG )
    """
    
    def __init__(self, group=None, combination=None, shuffle=False):
        # set combination
        self.__combination = []
        # initialize
        super(MoveGeneratorCombinator, self).__init__(group=group) 
        # set path
        self.set_combination(path)
        # set randomize
        self.set_shuffle(shuffle)
        
    @property
    def shuffle(self):
        """ Get shuffle flag."""
        return self.__shuffle
        
    @property
    def combination(self):
        """ Get the combination list of MoveGenerator instances."""
        return self.__combination
        
    def check_group(self, group):
        """
        Checks the generator's group.
        This methods always returns True because normally all combination MoveGenerator instances groups are checked.
        This method must NOT be overloaded unless needed.
        
        :Parameters:
            #. group (Group): the Group instance
        """
        return True, ""
    
    def set_group(self, group):
        """
        Set the MoveGenerator group.
        
        :Parameters:
            #. group (None, Group): group instance. 
        """
        MoveGenerator.set_group(self, group)
        for mg in self.__combination:
            mg.set_group(group)
         
    def set_combination(self, combination):
        """
        Sets the generators combination list.
        
        :Parameters:
            #. combination (list): The list of MoveGenerator instances.
        """
        assert isinstance(combination, (list,set,tuple)), LOGGER.error("combination must be a list")
        combination = list(combination)
        for c in combination:
            assert isinstance(c, MoveGenerator), LOGGER.error("every item in combination list must be a MoveGenerator instance")
            c.set_group(self.group)
        self.__combination = combination
        
    def set_shuffle(self, shuffle):
        """
        Sets whether to shuffle moves generator.
        
        :Parameters:
            #. shuffle (boolean): Whether to shuffle generator instances at every move or to combine moves in the list order.
        """
        assert isinstance(shuffle, bool), LOGGER.error("shuffle must be boolean")
        self.__shuffle = shuffle

    def move(self, coordinates):
        """
        Moves coordinates.
        
        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply the transformation
            
        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying the transformation
        """
        indexes = range(len(self.__combination))
        if self.__shuffle:
            shuffle( indexes )
        # create the move combination
        for idx in indexes:
            coordinates = self.__combination[idx].move(coordinates)
        return coordinates
    
        
class MoveGeneratorCollector(MoveGenerator):
    """ 
    MoveGeneratorCollector collects MoveGenerators instances and applies the move of one instance at every step.
    
    :Parameters:
        #. group (None, Group): The constraint RMC engine.
        #. collection (list): The list of MoveGenerator instances.
        #. randomize (boolean): Whether to pull MoveGenerator instance randomly from collection list or in order.     
        #. weights (None, list): Generators selections Weights list. 
           It must be None for equivalent weighting or list of (generatorIndex, weight) tuples. 
           If randomize is False, weights list is ignored upon generator selection from collection.
    
    .. code-block:: python
    
        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Core.MoveGenerator import MoveGeneratorCollector
        from fullrmc.Generators.Translations import TranslationGenerator
        from fullrmc.Generators.Rotations import RotationGenerator
        
        # create engine 
        ENGINE = Engine(pdb='system.pdb')
        
        # Add constraints ...
        # Re-define groups if needed ...
        
        ##### Define each group move generator as a combination of a translation and a rotation. #####
        # create recursive group selector. Recurrence is set to 20 with explore flag set to True.
        # randomize is set to True which means that at every selection a generator is randomly
        # selected from the collector.
        for g in ENGINE.groups:
            # create translation generator
            TMG = TranslationGenerator(amplitude=0.2)
            # create rotation generator only when group length is bigger than 1.
            if len(g)>1:
                RMG = RotationGenerator(amplitude=2)
                MG  = MoveGeneratorCollector(collection=[TMG,RMG],randomize=True)
            else:
                MG  = MoveGeneratorCollector(collection=[TMG],randomize=True)
            g.set_move_generator( MG )
            
    """
    def __init__(self, group=None, collection=None, randomize=True, weights=None):
        # set collection
        self.__collection = []
        # initialize
        super(MoveGeneratorCollector, self).__init__(group=group) 
        # set path
        self.set_collection(collection)
        # set randomize
        self.set_randomize(randomize)
        # set weights
        self.set_weights(weights)
        # initialize flags
        self.__initialize_generator()
    
    def __initialize_generator(self):
        self.__step = 0
    
    def __check_single_weight(self, w):
        """Checks a single group weight tuple format"""
        assert isinstance(w, (list,set,tuple)),LOGGER.error("weights list items must be tuples")
        assert len(w)==2, LOGGER.error("weights list tuples must have exactly 2 items")
        idx  = w[0]
        wgt = w[1]
        assert is_integer(idx), LOGGER.error("weights list tuples first item must be an integer")
        idx = INT_TYPE(idx)
        assert idx>=0, LOGGER.error("weights list tuples first item must be positive")
        assert idx<len(self.__collection), LOGGER.error("weights list tuples first item must be smaller than the number of generators in collection")
        assert is_number(wgt), LOGGER.error("weights list tuples second item must be an integer")
        wgt = FLOAT_TYPE(wgt)
        assert wgt>0, LOGGER.error("weights list tuples first item must be bigger than 0")
        # all True return idx and weight
        return idx, wgt  
        
    @property
    def randomize(self):
        """ Get randomize flag."""
        return self.__randomize
        
    @property
    def collection(self):
        """ Get the list of MoveGenerator instances."""
        return self.__collection
        
    @property
    def generatorsWeight(self):
        """ Generators selection weights list."""
        return self.__generatorsWeight
        
    @property    
    def selectionScheme(self):
        return self.__selectionScheme
    
    def set_group(self, group):
        """
        Set the MoveGenerator group.
        
        :Parameters:
            #. group (None, Group): group instance. 
        """
        MoveGenerator.set_group(self, group)
        for mg in self.__collection:
            mg.set_group(group)
            
    def check_group(self, group):
        """
        Checks the generator's group.
        This methods always returns True because normally all collection MoveGenerator instances groups are checked.
        This method must NOT be overloaded unless needed.
        
        :Parameters:
            #. group (Group): the Group instance.
        """
        return True, ""
        
    def set_collection(self, collection):
        """
        Sets the generators instances collection list.
        
        :Parameters:
            #. collection (list): The list of move generator instance.
        """
        assert isinstance(collection, (list,set,tuple)), LOGGER.error("collection must be a list")
        collection = list(collection)
        for c in collection:
            assert isinstance(c, MoveGenerator), LOGGER.error("every item in collection list must be a MoveGenerator instance")
            c.set_group(self.group)
        self.__collection = collection
        # reset generator
        self.__initialize_generator()
        
    def set_randomize(self, randomize):
        """
        Sets whether to randomize MoveGenerator instance selection from collection list.
        
        :Parameters:
            #. randomize (boolean): Whether to pull MoveGenerator instance randomly from collection list or in order.
        """
        assert isinstance(randomize, bool), LOGGER.error("randomize must be boolean")
        self.__randomize = randomize

    def set_weights(self, weights): 
        """
        Set groups selection weighting scheme.
        
        :Parameters:
            #. weights (None, list): Generators selections Weights list. 
               It must be None for equivalent weighting or list of (generatorIndex, weight) tuples. 
               If randomize is False, weights list is ignored upon generator selection from collection.
        """  
        generatorsWeight = np.ones(len(self.__collection), dtype=FLOAT_TYPE)      
        if weights is not None:
            assert isinstance(weights, (list,set,tuple)),LOGGER.error("weights must be a list")
            for w in weights:
                idx, wgt = self.__check_single_weight(w)
                # update groups weight
                generatorsWeight[idx] = wgt
        # set groups weight
        self.__generatorsWeight = generatorsWeight
        # create selection histogram
        self.set_selection_scheme()
    
    def set_selection_scheme(self):
        """ Sets selection scheme. """
        cumsumWeights = np.cumsum(self.__generatorsWeight, dtype=FLOAT_TYPE)
        self.__selectionScheme = cumsumWeights/cumsumWeights[-1]
        
    def move(self, coordinates):
        """
        Moves coordinates.
        
        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply the transformation
            
        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying the transformation
        """
        if self.__randomize:
            index = INT_TYPE( np.searchsorted(self.__selectionScheme, generate_random_float()) )
            moveGenerator = self.__collection[ index ]
        else:
            moveGenerator = self.__collection[self.__step]
            self.__step   = (self.__step+1)%len(self.__collection)
        # perform the move
        return moveGenerator.move(coordinates) 
        
                
        
        


"""
AtomicCoordinationConstraints contains classes for all constraints related to coordination numbers in shells around atoms.

.. inheritance-diagram:: fullrmc.Constraints.AtomicCoordinationConstraints
    :parts: 1
    
"""

# standard libraries imports
import itertools
import copy

# external libraries imports
import numpy as np

# fullrmc imports
from fullrmc.Globals import INT_TYPE, FLOAT_TYPE, PI, PRECISION, FLOAT_PLUS_INFINITY, LOGGER
from fullrmc.Core.Collection import is_number, is_integer
from fullrmc.Core.Constraint import Constraint, SingularConstraint, RigidConstraint
from fullrmc.Core.atomic_coordination import all_atoms_coord_number_coords, multi_atoms_coord_number_coords


class AtomicCoordinationNumberConstraint(RigidConstraint, SingularConstraint):
    """
    It's a rigid constraint that controls the coordination number between atoms. 
    
    .. raw:: html

        <iframe width="560" height="315" 
        src="https://www.youtube.com/embed/R8t-_XwizOI?rel=0" 
        frameborder="0" allowfullscreen>
        </iframe>
        
        
    :Parameters:
        #. engine (None, fullrmc.Engine): The constraint RMC engine.
        #. coordNumDef (None, list, tuple): The coordination number definition. 
           It must be None or a list or tuple where every element is a list or a
           tuple of exactly 6 items and an optional 7th item for weight.
           
           #. the core atoms: Can be any of the following:
           
              * string: indicating atomic element
              * dictionary: Key an atomic attribute among (element, name) 
                and value is the attribute value.
              * list, tuple, set, numpy.ndarray: core atoms indexes  
           
           #. the in shell atoms: Can be any of the following:
           
              * string: indicating atomic element
              * dictionary: Key an atomic attribute among (element, name) 
                and value is the attribute value.
              * list, tuple, set, numpy.ndarray: in shell atoms indexes  
           
           #. the lower distance limit of the coordination shell.
           #. the upper distance limit of the coordination shell.
           #. :math:`N_{min}` : the minimum number of neighbours in the shell.
           #. :math:`N_{max}` : the maximum number of neighbours in the shell.
           #. :math:`W_{i}` : the weight contribution to the standard error, 
              this is optional, if not given it is set automatically to 1.0.
        #. rejectProbability (Number): rejecting probability of all steps where standardError increases. 
           It must be between 0 and 1 where 1 means rejecting all steps where standardError increases
           and 0 means accepting all steps regardless whether standardError increases or not.
    """
    def __init__(self, engine, coordNumDef=None, rejectProbability=1):
        # initialize constraint
        RigidConstraint.__init__(self, engine=engine, rejectProbability=rejectProbability)
        # initialize data
        self.__initialize_constraint_data()
        # set coordination number definition
        self.set_coordination_number_definition(coordNumDef)
        # set computation cost
        self.set_computation_cost(5.0)
    
    def __initialize_constraint_data(self):
        # the following is the parsing of defined shells
        self.__coresIndexes  = []
        self.__shellsIndexes = []
        self.__lowerShells   = []
        self.__upperShells   = []
        self.__minAtoms      = []
        self.__maxAtoms      = []
        # upon computing constraint data, those values must be divided by len( self.__coresIndexes[i] )
        self.__coordNumData = []
        self.__weights      = [] 
        # atoms to cores and shells pointers
        self.__asCoreDefIdxs  = []
        self.__inShellDefIdxs = []
        
    @property
    def coordNumDef(self):
        """Get coordination number definition dictionary"""
        return self.__coordNumDef
    
    @property
    def coresIndexes(self):  
        """Get the list of coordination number core atoms indexes array as generated 
        from  coordination number definition."""  
        return self.__coresIndexes
        
    @property
    def shellsIndexes(self):  
        """ Get the list of coordination number shell atoms indexes array as generated 
        from coordination number definition."""
        return self.__shellsIndexes
        
    @property
    def lowerShells(self):  
        """Get array of lower shells distance as generated from coordination number 
        definition. """  
        return self.__lowerShells
    
    @property
    def upperShells(self):  
        """Get array of upper shells distance as generated from coordination number 
        definition. """    
        return self.__upperShells
    
    @property
    def minAtoms(self):  
        """Get array of minimum number of atoms in a shell as generated from 
        coordination number definition. """   
        return self.__minAtoms
    
    @property
    def maxAtoms(self):  
        """Get array of maximum number of atoms in a shell as generated from 
        coordination number definition. """
        return self.__maxAtoms
    
    @property
    def weights(self):  
        """Get shells weights which count in the computation of standard error."""  
        return self.__weights
        
    @property
    def data(self):  
        """Get coordination number constraint data."""  
        return self.__coordNumData
    
    @property
    def asCoreDefIdxs(self):  
        """Get the list of arrays where each element is pointing to a coordination 
        number definition where the atom is a core."""  
        return self.__asCoreDefIdxs
    
    @property
    def inShellDefIdxs(self):  
        """Get the list of arrays where each element is pointing to a coordination 
        number definition where the atom is in a shell."""    
        return self.__inShellDefIdxs
               
    def listen(self, message, argument=None):
        """   
        listen to any message sent from the Broadcaster.
        
        :Parameters:
            #. message (object): Any python object to send to constraint's listen method.
            #. argument (object): Any type of argument to pass to the listeners.
        """
        if message in("engine changed", "update molecules indexes"):
            self.set_coordination_number_definition(self.__coordNumDef)
        elif message in("update boundary conditions",):
            self.reset_constraint()        
    
    def set_coordination_number_definition(self, coordNumDef):
        """
        Set the coordination number definition.

        :Parameters:
            #. coordNumDef (None, list, tuple): The coordination number definition. 
               It must be None or a list or tuple where every element is a list or a
               tuple of exactly 6 items and an optional 7th item for weight.
               
               #. the core atoms: Can be any of the following:
               
                  * string: indicating atomic element
                  * dictionary: Key an atomic attribute among (element, name) 
                    and value is the attribute value.
                  * list, tuple, set, numpy.ndarray: core atoms indexes  
               
               #. the in shell atoms: Can be any of the following:
               
                  * string: indicating atomic element
                  * dictionary: Key an atomic attribute among (element, name) 
                    and value is the attribute value.
                  * list, tuple, set, numpy.ndarray: in shell atoms indexes  
               
               #. the lower distance limit of the coordination shell.
               #. the upper distance limit of the coordination shell.
               #. :math:`N_{min}` : the minimum number of neighbours in the shell.
               #. :math:`N_{max}` : the maximum number of neighbours in the shell.
               #. :math:`W_{i}` : the weight contribution to the standard error, 
                  this is optional, if not given it is set automatically to 1.0.
               
               ::

                   e.g. [ ('Ti','Ti', 2.5, 3.5, 5, 7.1, 1), ('Ni','Ti', 2.2, 3.1, 7.2, 9.7, 100), ...]
                        [ ({'element':'Ti'},'Ti', 2.5, 3.5, 5, 7.1, 0.1), ...]  
                        [ ({'name':'au'},'Au', 2.5, 3.5, 4.1, 6.3), ...]  
                        [ ({'name':'Ni'},{'element':'Ti'}, 2.2, 3.1, 7, 9), ...]   
                        [ ('Ti',range(100,500), 2.2, 3.1, 7, 9), ...]   
                        [ ([0,10,11,15,1000],{'name':'Ti'}, 2.2, 3.1, 7, 9, 5), ...]   
        
        """
        self.__coordNumDef = coordNumDef
        if self.engine is None:
            return
        elif coordNumDef is None:
            coordNumDef = []
        ########## check definitions, create coordination number data ########## 
        self.__initialize_constraint_data()
        for CNDef in coordNumDef:
            assert isinstance(CNDef, (list, tuple)), LOGGER.error("coordNumDef item must be a list or a tuple")
            if len(CNDef) == 6:
                coreDef, shellDef, lowerShell, upperShell, minCN, maxCN = CNDef
                weight = 1.0
            elif len(CNDef) == 7:
                coreDef, shellDef, lowerShell, upperShell, minCN, maxCN, weight = CNDef
            else:
                raise LOGGER.error("coordNumDef item must have 6 or 7 items")
            # core definition
            if isinstance(coreDef, basestring):
                coreDef = str(coreDef)
                assert coreDef in self.engine.elements, LOGGER.error("core atom definition '%s' is not a valid element"%coreDef)
                coreIndexes = [idx for idx, el in enumerate(self.engine.allElements) if el==coreDef]
            elif isnstance(coreDef, dict):
                assert len(coreDef) == 1, LOGGER.error("core atom definition dictionary must be of length 1")
                key, value = coreDef.keys()[0], coreDef.values()[0]
                if key is "name":
                    assert value in self.engine.names, LOGGER.error("core atom definition '%s' is not a valid name"%coreDef)
                    coreIndexes = [idx for idx, el in enumerate(self.engine.allNames) if el==coreDef]
                elif key is "element":
                    assert value in self.engine.elements, LOGGER.error("core atom definition '%s' is not a valid element"%coreDef)
                    coreIndexes = [idx for idx, el in enumerate(self.engine.allElements) if el==coreDef]
                else:
                    raise LOGGER.error("core atom definition dictionary key must be either 'name' or 'element'")
            elif isnstance(coreDef, (list, tuple, set, np.ndarray)):
                coreIndexes = []
                if isinstance(coreDef, np.ndarray):
                    assert len(coreDef.shape)==1, LOGGER.error("core atom definition numpy.ndarray must be 1D")
                for c in coreDef:
                    assert is_integer(c), LOGGER.error("core atom definition index must be integer")
                    c = INT_TYPE(c)
                    assert c>=0, LOGGER.error("core atom definition index must be >=0")
                    assert c<len(self.engine.pdb), LOGGER.error("core atom definition index must be smaler than number of atoms in system")
                    coreIndexes.append(c)
            # shell definition
            if isinstance(shellDef, basestring):
                shellDef = str(shellDef)
                assert shellDef in self.engine.elements, LOGGER.error("core atom definition '%s' is not a valid element"%shellDef)
                shellIndexes = [idx for idx, el in enumerate(self.engine.allElements) if el==shellDef]
            elif isnstance(shellDef, dict):
                assert len(shellDef) == 1, LOGGER.error("core atom definition dictionary must be of length 1")
                key, value = shellDef.keys()[0], shellDef.values()[0]
                if key is "name":
                    assert value in self.engine.names, LOGGER.error("core atom definition '%s' is not a valid name"%shellDef)
                    shellIndexes = [idx for idx, el in enumerate(self.engine.allNames) if el==shellDef]
                elif key is "element":
                    assert value in self.engine.elements, LOGGER.error("core atom definition '%s' is not a valid element"%shellDef)
                    shellIndexes = [idx for idx, el in enumerate(self.engine.allElements) if el==shellDef]
                else:
                    raise LOGGER.error("core atom definition dictionary key must be either 'name' or 'element'")
            elif isnstance(shellDef, (list, tuple, set, np.ndarray)):
                shellIndexes = []
                if isinstance(shellDef, np.ndarray):
                    assert len(shellDef.shape)==1, LOGGER.error("core atom definition numpy.ndarray must be 1D")
                for c in shellDef:
                    assert is_integer(c), LOGGER.error("core atom definition index must be integer")
                    c = INT_TYPE(c)
                    assert c>=0, LOGGER.error("core atom definition index must be >=0")
                    assert c<len(self.engine.pdb), LOGGER.error("core atom definition index must be smaler than number of atoms in system")
                    shellIndexes.append(c)
            # lower and upper shells definition
            assert is_number(lowerShell), LOGGER.error("Coordination number lower shell '%s' must be a number."%lowerShell)       
            lowerShell = FLOAT_TYPE(lowerShell)
            assert lowerShell>=0, LOGGER.error("Coordination number lower shell '%s' must be a positive."%lowerShell)       
            assert is_number(upperShell), LOGGER.error("Coordination number upper shell '%s' must be a number."%key)       
            upperShell = FLOAT_TYPE(upperShell)
            assert upperShell>lowerShell, LOGGER.error("Coordination number lower shell '%s' must be smaller than upper shell %s"%(lowerShell,upperShell))       
            # minimum and maximum number of atoms definitions
            assert is_number(minCN), LOGGER.error("Coordination number minimum atoms '%s' must be a number."%minCN)       
            minCN = FLOAT_TYPE(minCN)
            assert minCN>=0, LOGGER.error("Coordination number minimim atoms '%s' must be >=0."%minCN)       
            assert is_number(maxCN), LOGGER.error("Coordination number maximum atoms '%s' must be a number."%key)       
            maxCN = FLOAT_TYPE(maxCN)
            assert maxCN>=minCN, LOGGER.error("Coordination number minimum atoms '%s' must be smaller than maximum atoms %s"%(minCN,maxCN))       
            # check weight
            assert is_number(weight), LOGGER.error("Coordination number weight '%s' must be a number."%weight)       
            weight = FLOAT_TYPE(weight)
            assert weight>0, LOGGER.error("Coordination number weight '%s' must be >0."%weight)       
            # append coordination number data
            self.__coresIndexes.append( sorted(set(coreIndexes)) )
            self.__shellsIndexes.append( sorted(set(shellIndexes)) ) 
            self.__lowerShells.append( lowerShell )    
            self.__upperShells.append( upperShell )    
            self.__minAtoms.append( minCN )      
            self.__maxAtoms.append( maxCN ) 
            self.__coordNumData.append( FLOAT_TYPE(0) ) 
            self.__weights.append( weight ) 
        ########## set asCoreDefIdxs and inShellDefIdxs points ##########  
        for _ in xrange(self.engine.numberOfAtoms):
            self.__asCoreDefIdxs.append( [] )
            self.__inShellDefIdxs.append( [] )
        for defIdx, indexes in enumerate(self.__coresIndexes):
            self.__coresIndexes[defIdx] = np.array( indexes, dtype=INT_TYPE )
            for atIdx in indexes:
                self.__asCoreDefIdxs[atIdx].append( defIdx )
        for defIdx, indexes in enumerate(self.__shellsIndexes):
            self.__shellsIndexes[defIdx] = np.array( indexes, dtype=INT_TYPE )
            for atIdx in indexes:
                self.__inShellDefIdxs[atIdx].append( defIdx )
        for atIdx in xrange(self.engine.numberOfAtoms):
            self.__asCoreDefIdxs[atIdx]  = np.array( self.__asCoreDefIdxs[atIdx], dtype=INT_TYPE )
            self.__inShellDefIdxs[atIdx] = np.array( self.__inShellDefIdxs[atIdx], dtype=INT_TYPE )
        # set all to arrays
        self.__coordNumData  = np.array( self.__coordNumData, dtype=FLOAT_TYPE )
        self.__weights       = np.array( self.__weights, dtype=FLOAT_TYPE )
        self.__numberOfCores = np.array( [len(idxs) for idxs in self.__coresIndexes], dtype=FLOAT_TYPE )

    def compute_standard_error(self, data):
        """ 
        Compute the standard error (StdErr) of data not satisfying constraint conditions. 
        
        .. math::        
            StdErr = \\sum \\limits_{i}^{S} Dev_{i}


        .. math::
            Dev_{i}=\\begin{cases}
              W_{i}*( N_{min,i}-\\overline{CN_{i}} ), & \\text{if $\\overline{CN_{i}}<N_{min,i}$}.\\\\
              W_{i}*( \\overline{CN_{i}}-N_{max,i} ), & \\text{if $\\overline{CN_{i}}>N_{max,i}$}.\\\\
              0                  , & \\text{if $N_{min,i}<=\\overline{CN_{i}}<=N_{max,i}$}
            \\end{cases}
        
                
        Where:\n
        :math:`S`                  is the total number of defined coordination number shells. \n
        :math:`W_{i}`              is the defined weight of coordination number shell i. \n
        :math:`Dev_{i}`            is the standard deviation of the coordination number in shell definition i. \n
        :math:`\\overline{CN_{i}}` is the mean coordination number value in shell definition i. \n
        :math:`N_{min,i}`          is the defined minimum number of neighbours in shell definition i. \n
        :math:`N_{max,i}`          is the defined maximum number of neighbours in shell definition i. \n
         

        :Parameters:
            #. data (numpy.array): The constraint value data to compute standardError.
            
        :Returns:
            #. standardError (number): The calculated standardError of the constraint.
        """
        coordNum = data/self.__numberOfCores
        StdErr   = 0.
        for idx, cn in enumerate( coordNum ):
            if cn < self.__minAtoms[idx]:
                StdErr += self.__weights[idx]*(self.__minAtoms[idx]-cn)
            elif cn > self.__maxAtoms[idx]:
                StdErr += self.__weights[idx]*(cn-self.__maxAtoms[idx])
        return StdErr
        
    def get_constraint_value(self):
        """
        Gets squared deviation per shell definition
        
        :Returns:
            #. MPD (dictionary): The MPD dictionary, where keys are the element wise intra and inter molecular MPDs and values are the computed MPDs.
        """
        if self.data is None:
            log.LocalLogger("fullrmc").logger.warn("data must be computed first using 'compute_data' method.")
            return None
        
    def compute_data(self):
        """ Compute data and update engine constraintsData dictionary. """
        all_atoms_coord_number_coords(boxCoords      = self.engine.boxCoordinates,
                                      basis          = self.engine.basisVectors,
                                      isPBC          = self.engine.isPBC,
                                      coresIndexes   = self.__coresIndexes,
                                      shellsIndexes  = self.__shellsIndexes,
                                      lowerShells    = self.__lowerShells,
                                      upperShells    = self.__upperShells,
                                      asCoreDefIdxs  = self.__asCoreDefIdxs,
                                      inShellDefIdxs = self.__inShellDefIdxs,
                                      coordNumData   = self.__coordNumData,
                                      ncores         = self.engine._runtime_ncores)
        self.__coordNumData /= FLOAT_TYPE(2.)         
        # update data
        self.set_data( self.__coordNumData )
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # set standardError
        stdErr = self.compute_standard_error(data = self.__coordNumData)
        self.set_standard_error(stdErr)

    def compute_before_move(self, indexes):
        """ 
        Compute constraint before move is executed
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        beforeMoveData = np.zeros(self.__coordNumData.shape, dtype=self.__coordNumData.dtype)
        multi_atoms_coord_number_coords( indexes        = indexes,
                                         boxCoords      = self.engine.boxCoordinates,
                                         basis          = self.engine.basisVectors,
                                         isPBC          = self.engine.isPBC,
                                         coresIndexes   = self.__coresIndexes,
                                         shellsIndexes  = self.__shellsIndexes,
                                         lowerShells    = self.__lowerShells,
                                         upperShells    = self.__upperShells,
                                         asCoreDefIdxs  = self.__asCoreDefIdxs,
                                         inShellDefIdxs = self.__inShellDefIdxs,
                                         coordNumData   = beforeMoveData,
                                         ncores         = self.engine._runtime_ncores)
        # set active atoms data before move
        self.set_active_atoms_data_before_move( beforeMoveData )
        self.set_active_atoms_data_after_move(None)                                                   
           
    def compute_after_move(self, indexes, movedBoxCoordinates):
        """ 
        Compute constraint after move is executed
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to.
            #. movedBoxCoordinates (numpy.ndarray): The moved atoms new coordinates.
        """
        # change coordinates temporarily
        boxData = np.array(self.engine.boxCoordinates[indexes], dtype=FLOAT_TYPE)
        self.engine.boxCoordinates[indexes] = movedBoxCoordinates
        # compute after move data
        afterMoveData = np.zeros(self.__coordNumData.shape, dtype=self.__coordNumData.dtype)
        multi_atoms_coord_number_coords( indexes        = indexes,
                                         boxCoords      = self.engine.boxCoordinates,
                                         basis          = self.engine.basisVectors,
                                         isPBC          = self.engine.isPBC,
                                         coresIndexes   = self.__coresIndexes,
                                         shellsIndexes  = self.__shellsIndexes,
                                         lowerShells    = self.__lowerShells,
                                         upperShells    = self.__upperShells,
                                         asCoreDefIdxs  = self.__asCoreDefIdxs,
                                         inShellDefIdxs = self.__inShellDefIdxs,
                                         coordNumData   = afterMoveData,
                                         ncores         = self.engine._runtime_ncores)
        # reset coordinates
        self.engine.boxCoordinates[indexes] = boxData
        # set active atoms data after move
        self.set_active_atoms_data_after_move( afterMoveData )
        # compute after move standard error
        self.__coordNumDataAfterMove = self.__coordNumData-self.activeAtomsDataBeforeMove+self.activeAtomsDataAfterMove
        self.set_after_move_standard_error( self.compute_standard_error(data = self.__coordNumDataAfterMove) )

    def accept_move(self, indexes):
        """ 
        Accept move.
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        self.__coordNumData = self.__coordNumDataAfterMove
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # update standardError
        self.set_standard_error(self.afterMoveStandardError)
        self.set_after_move_standard_error( None )
        
    def reject_move(self, indexes):
        """ 
        Reject move.
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # update standardError
        self.set_after_move_standard_error( None )


    
    
            


"""
ImproperAngleConstraints contains classes for all constraints related improper angles between atoms.

.. inheritance-diagram:: fullrmc.Constraints.ImproperAngleConstraints
    :parts: 1
"""

# standard libraries imports
import itertools

# external libraries imports
import numpy as np

# fullrmc imports
from fullrmc.Globals import INT_TYPE, FLOAT_TYPE, PI, PRECISION, FLOAT_PLUS_INFINITY, LOGGER
from fullrmc.Core.Collection import is_number, is_integer, get_path
from fullrmc.Core.Constraint import Constraint, SingularConstraint, RigidConstraint
from fullrmc.Core.improper_angles import full_improper_angles_coords

class ImproperAngleConstraint(RigidConstraint, SingularConstraint):
    """
    Controls the improper angle formed with 4 defined atoms. It's mainly used to 
    keep the improper atom in the plane defined with three other atoms. 
    The improper vector is defined as the vector from the first atom of the plane to
    the improper atom. Therefore the improper angle is defined between the improper
    vector and the plane.
    
     .. raw:: html

        <iframe width="560" height="315" 
        src="https://www.youtube.com/embed/qVATE-9cIBg" 
        frameborder="0" allowfullscreen>
        </iframe> 
        
        
    :Parameters:
        #. engine (None, fullrmc.Engine): The constraint RMC engine.
        #. anglesMap (list): The angles map definition.
               Every item must be a list of five items.
               
               #. First item: The improper atom index that must be in the plane.
               #. Second item: The index of the atom 'O' considered the origin of the plane.
               #. Third item: The index of the atom 'x' used to calculated 'Ox' vector.
               #. Fourth item: The index of the atom 'y' used to calculated 'Oy' vector.
               #. Fifth item: The minimum lower limit or the minimum angle allowed 
                  in degrees which later will be converted to rad.
               #. Sixth item: The maximum upper limit or the maximum angle allowed 
                  in degrees which later will be converted to rad.
        #. rejectProbability (Number): rejecting probability of all steps where standardError increases. 
           It must be between 0 and 1 where 1 means rejecting all steps where standardError increases
           and 0 means accepting all steps regardless whether standardError increases or not.
    
    .. code-block:: python
        
        ## Tetrahydrofuran (THF) molecule sketch
        ## 
        ##              O
        ##   H41      /   \      H11
        ##      \  /         \  /
        ## H42-- C4    THF    C1 --H12
        ##        \  MOLECULE /
        ##         \         /
        ##   H31-- C3-------C2 --H21
        ##        /         \\
        ##     H32            H22 
        ##
 
        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Constraints.ImproperAngleConstraints import ImproperAngleConstraint
        
        # create engine 
        ENGINE = Engine(pdb='system.pdb')
        
        # create and add constraint
        IAC = ImproperAngleConstraint(engine=None)
        ENGINE.add_constraints(IAC)
        
        # define intra-molecular improper angles 
        IAC.create_angles_by_definition( anglesDefinition={"THF": [ ('C2','O','C1','C4', -15, 15),
                                                                    ('C3','O','C1','C4', -15, 15) ] })
           
                                                            
    """
    
    def __init__(self, engine, anglesMap=None, rejectProbability=1):
        # initialize constraint
        RigidConstraint.__init__(self, engine=engine, rejectProbability=rejectProbability)
        # set bonds map
        self.set_angles(anglesMap)
        # set computation cost
        self.set_computation_cost(3.0)
        
    @property
    def anglesMap(self):
        """ Get angles map."""
        return self.self.__anglesMap
    
    @property
    def angles(self):
        """ Get angles dictionary."""
        return self.__angles
    
    @property
    def atomsLUAD(self):
        """ 
        Get look up angles dictionary, connecting every atom's index to a central 
        atom angles definition of angles attribute.
        """
        return self.__atomsLUAD
        
    @property
    def standardError(self):
        """ Get constraint's current squared deviation."""
        if self.data is None:
            return None
        else: 
            return self.compute_standard_error(data = self.data)
            
    def listen(self, message, argument=None):
        """   
        Listens to any message sent from the Broadcaster.
        
        :Parameters:
            #. message (object): Any python object to send to constraint's listen method.
            #. argument (object): Any type of argument to pass to the listeners.
        """
        if message in("engine changed","update boundary conditions",):
            self.reset_constraint()        
        
    def should_step_get_rejected(self, standardError):
        """
        Overloads 'RigidConstraint' should_step_get_rejected method.
        It computes whether to accept or reject a move based on before and after move calculation and not standardError.
        If any of activeAtomsDataBeforeMove or activeAtomsDataAfterMove is None an Exception will get raised.
        
        :Parameters:
            #. standardError (number): not used in this case
        
        :Return:
            #. result (boolean): True to reject step, False to accept
        """
        if self.activeAtomsDataBeforeMove is None or self.activeAtomsDataAfterMove is None:
            raise Exception(LOGGER.error("must compute data before and after group move"))
        reject = False
        for index in self.activeAtomsDataBeforeMove.keys():
            before = self.activeAtomsDataBeforeMove[index]["reducedAngles"]
            after  = self.activeAtomsDataAfterMove[index]["reducedAngles"]
            if np.any((after-before)>PRECISION):
                reject = True
                break
        return reject
        
    def set_angles(self, anglesMap):
        """ 
        Sets the angles dictionary by parsing the anglesMap list.
        
        :Parameters:
            #. anglesMap (list): The angles map definition.
               Every item must be a list of five items.\n
               #. First item: The improper atom index that must be in the plane.
               #. Second item: The index of the atom 'O' considered the origin of the plane.
               #. Third item: The index of the atom 'x' used to calculated 'Ox' vector.
               #. Fourth item: The index of the atom 'y' used to calculated 'Oy' vector.
               #. Fifth item: The minimum lower limit or the minimum angle allowed 
                  in degrees which later will be converted to rad.
               #. Sixth item: The maximum upper limit or the maximum angle allowed 
                  in degrees which later will be converted to rad.
        """
        map = []
        if self.engine is not None:
            if anglesMap is not None:
                assert isinstance(anglesMap, (list, set, tuple)), LOGGER.error("anglesMap must be None or a list")
                for angle in anglesMap:
                    assert isinstance(angle, (list, set, tuple)), LOGGER.error("anglesMap items must be lists")
                    angle = list(angle)
                    assert len(angle)==6, LOGGER.error("anglesMap items must be lists of 6 items each")
                    improperIdx, oIdx, xIdx, yIdx, lower, upper = angle
                    assert is_integer(improperIdx), LOGGER.error("anglesMap items lists of first item must be an integer")
                    improperIdx = INT_TYPE(improperIdx)
                    assert is_integer(oIdx), LOGGER.error("anglesMap items lists of second item must be an integer")
                    oIdx = INT_TYPE(oIdx)
                    assert is_integer(xIdx), LOGGER.error("anglesMap items lists of third item must be an integer")
                    xIdx = INT_TYPE(xIdx)
                    assert is_integer(yIdx), LOGGER.error("anglesMap items lists of fourth item must be an integer")
                    yIdx = INT_TYPE(yIdx)
                    assert improperIdx>=0, LOGGER.error("anglesMap items lists first item must be positive")
                    assert oIdx>=0, LOGGER.error("anglesMap items lists second item must be positive")
                    assert xIdx>=0, LOGGER.error("anglesMap items lists third item must be positive")
                    assert yIdx>=0, LOGGER.error("anglesMap items lists fourth item must be positive")
                    assert improperIdx!=oIdx, LOGGER.error("bondsMap items lists first and second items can't be the same")
                    assert improperIdx!=xIdx, LOGGER.error("bondsMap items lists first and third items can't be the same")
                    assert improperIdx!=yIdx, LOGGER.error("bondsMap items lists first and fourth items can't be the same")
                    assert oIdx!=xIdx, LOGGER.error("bondsMap items lists second and third items can't be the same")
                    assert oIdx!=yIdx, LOGGER.error("bondsMap items lists second and fourth items can't be the same")
                    assert xIdx!=yIdx, LOGGER.error("bondsMap items lists third and fourth items can't be the same")
                    assert is_number(lower), LOGGER.error("anglesMap items lists of fifth item must be a number")
                    lower = FLOAT_TYPE(lower)
                    assert is_number(upper), LOGGER.error("anglesMap items lists of sixth item must be a number")
                    upper = FLOAT_TYPE(upper)
                    assert lower>=-90, LOGGER.error("anglesMap items lists fifth item must be bigger or equal to -90 deg.")
                    assert upper>lower, LOGGER.error("anglesMap items lists fifth item must be smaller than the sixth item")
                    assert upper<=90, LOGGER.error("anglesMap items lists sixth item must be smaller or equal to 90")
                    lower *= FLOAT_TYPE( PI/FLOAT_TYPE(180.) )
                    upper *= FLOAT_TYPE( PI/FLOAT_TYPE(180.) )
                    map.append((improperIdx, oIdx, xIdx, yIdx, lower, upper))  
        # set anglesMap definition
        self.__anglesMap = map      
        # create bonds list of indexes arrays
        self.__angles = {}
        self.__atomsLUAD = {}
        if self.engine is not None:
            # parse bondsMap
            for angle in self.__anglesMap:
                improperIdx, oIdx, xIdx, yIdx, lower, upper = angle
                assert improperIdx<len(self.engine.pdb), LOGGER.error("angle atom index must be smaller than maximum number of atoms")
                assert oIdx<len(self.engine.pdb), LOGGER.error("angle atom index must be smaller than maximum number of atoms")
                assert xIdx<len(self.engine.pdb), LOGGER.error("angle atom index must be smaller than maximum number of atoms")
                assert yIdx<len(self.engine.pdb), LOGGER.error("angle atom index must be smaller than maximum number of atoms")
                # create atoms look up angles dictionary
                if not self.__atomsLUAD.has_key(improperIdx):
                    self.__atomsLUAD[improperIdx] = []
                if not self.__atomsLUAD.has_key(oIdx):
                    self.__atomsLUAD[oIdx] = []
                if not self.__atomsLUAD.has_key(xIdx):
                    self.__atomsLUAD[xIdx] = []
                if not self.__atomsLUAD.has_key(yIdx):
                    self.__atomsLUAD[yIdx] = []
                # create improper angles
                if not self.__angles.has_key(improperIdx):
                    self.__angles[improperIdx] = {"oIndexes":[],"xIndexes":[],"yIndexes":[],"lower":[],"upper":[]}
                # check for redundancy and append
                elif oIdx in self.__angles[improperIdx]["oIndexes"]:
                    index = self.__angles[improperIdx]["oIndexes"].index(oIndexes)
                    if sorted(oIdx,xIdx,yIdx) == sorted(self.__angles[improperIdx]["oIndexes"][index],self.__angles[improperIdx]["xIndexes"][index],self.__angles[improperIdx]["yIndexes"][index]):
                        log.LocalLogger("fullrmc").logger.warn("Improper angle definition for improper atom index '%i' and (O,x,y) atoms indexes (%i,%i,%i)  already defined. New angle limits [%.3f,%.3f] ignored and old angle limits [%.3f,%.3f] kept."%(improperIdx, oIdx, xIdx, yIdx, lower, upper, self.__angles[improperIdx]["lower"][index], self.__angles[improperIdx]["upper"][index]))
                        continue
                # add improper angle definition
                self.__angles[improperIdx]["oIndexes"].append(oIdx)
                self.__angles[improperIdx]["xIndexes"].append(xIdx)
                self.__angles[improperIdx]["yIndexes"].append(yIdx)
                self.__angles[improperIdx]["lower"].append(lower)
                self.__angles[improperIdx]["upper"].append(upper)
                self.__atomsLUAD[improperIdx].append(improperIdx)
                self.__atomsLUAD[oIdx].append(improperIdx)
                self.__atomsLUAD[xIdx].append(improperIdx)
                self.__atomsLUAD[yIdx].append(improperIdx)
            # finalize improper angles
            for idx in self.engine.pdb.xindexes:
                angles = self.__angles.get(idx, {"oIndexes":[],"xIndexes":[],"yIndexes":[],"lower":[],"upper":[]} )
                self.__angles[INT_TYPE(idx)] =  {"oIndexes": np.array(angles["oIndexes"], dtype = INT_TYPE), 
                                                 "xIndexes": np.array(angles["xIndexes"], dtype = INT_TYPE),
                                                 "yIndexes": np.array(angles["yIndexes"], dtype = INT_TYPE),
                                                 "lower"  : np.array(angles["lower"],     dtype = FLOAT_TYPE),
                                                 "upper"  : np.array(angles["upper"],     dtype = FLOAT_TYPE) }
                lut = self.__atomsLUAD.get(idx, [] )
                self.__atomsLUAD[INT_TYPE(idx)] = sorted(set(lut))
        # reset constraint
        self.reset_constraint()
    
    def create_angles_by_definition(self, anglesDefinition):
        """ 
        Creates anglesMap using angles definition.
        Calls set_angles(anglesMap) and generates angles attribute.
        
        :Parameters:
            #. anglesDefinition (dict): The angles definition. 
               Every key must be a molecule name (residue name in pdb file). 
               Every key value must be a list of angles definitions. 
               Every angle definition is a list of five items where:
               
               #. First item: The name of the improper atom that must be in the plane.
               #. Second item: The name of the atom 'O' considered the origin of the plane.
               #. Third item: The name of the atom 'x' used to calculated 'Ox' vector.
               #. Fourth item: The name of the atom 'y' used to calculated 'Oy' vector.
               #. Fifth item: The minimum lower limit or the minimum angle allowed 
                  in degrees which later will be converted to rad.
               #. Sixth item: The maximum upper limit or the maximum angle allowed 
                  in degrees which later will be converted to rad.
        
        ::
        
            e.g. (Benzene):  anglesDefinition={"BENZ": [('C3','C1','C2','C6', -10, 10),
                                                        ('C4','C1','C2','C6', -10, 10),
                                                        ('C5','C1','C2','C6', -10, 10) ] }
                                                  
        """
        if self.engine is None:
            raise Exception("Engine is not defined. Can't create angles")
        assert isinstance(anglesDefinition, dict), "anglesDefinition must be a dictionary"
        # check map definition
        existingMoleculesNames = sorted(set(self.engine.moleculesNames))
        anglesDef = {}
        for mol, angles in anglesDefinition.items():
            if mol not in existingMoleculesNames:
                log.LocalLogger("fullrmc").logger.warn("Molecule name '%s' in anglesDefinition is not recognized, angles definition for this particular molecule is omitted"%str(mol))
                continue
            assert isinstance(angles, (list, set, tuple)), LOGGER.error("mapDefinition molecule angles must be a list")
            angles = list(angles)
            molAnglesMap = []
            for angle in angles:
                assert isinstance(angle, (list, set, tuple)), LOGGER.error("mapDefinition angles must be a list")
                angle = list(angle)
                assert len(angle)==6
                improperAt, oAt, xAt, yAt, lower, upper = angle
                # check for redundancy
                append = True
                for b in molAnglesMap:
                    if (b[0]==improperAt):
                        if sorted([oAt,xAt,yAt]) == sorted([b[1],b[2],b[3]]):
                            log.LocalLogger("fullrmc").logger.warn("Improper angle definition for improper atom index '%i' and (O,x,y) atoms indexes (%i,%i,%i)  already defined. New angle limits [%.3f,%.3f] ignored and old angle limits [%.3f,%.3f] kept."%(improperIdx, oIdx, xIdx, yIdx, lower, upper, self.__angles[improperIdx]["lower"][index], self.__angles[improperIdx]["upper"][index]))
                            append = False
                            break
                if append:
                    molAnglesMap.append((improperAt, oAt, xAt, yAt, lower, upper))
            # create bondDef for molecule mol 
            anglesDef[mol] = molAnglesMap
        # create mols dictionary
        mols = {}
        for idx in self.engine.pdb.xindexes:
            molName = self.engine.moleculesNames[idx]
            if not molName in anglesDef.keys():    
                continue
            molIdx = self.engine.moleculesIndexes[idx]
            if not mols.has_key(molIdx):
                mols[molIdx] = {"name":molName, "indexes":[], "names":[]}
            mols[molIdx]["indexes"].append(idx)
            mols[molIdx]["names"].append(self.engine.allNames[idx])
        # get anglesMap
        anglesMap = []         
        for val in mols.values():
            indexes = val["indexes"]
            names   = val["names"]
            # get definition for this molecule
            thisDef = anglesDef[val["name"]]
            for angle in thisDef:
                improperIdx = indexes[ names.index(angle[0]) ]
                oIdx        = indexes[ names.index(angle[1]) ]
                xIdx        = indexes[ names.index(angle[2]) ]
                yIdx        = indexes[ names.index(angle[3]) ]
                lower       = angle[4]
                upper       = angle[5]
                anglesMap.append((improperIdx, oIdx, xIdx, yIdx, lower, upper))
        # create angles
        self.set_angles(anglesMap=anglesMap)
    
    def compute_standard_error(self, data):
        """ 
        Compute the standard error (StdErr) of data not satisfying constraint conditions. 
        
        .. math::
            StdErr = \\sum \\limits_{i}^{C} 
            ( \\theta_{i} - \\theta_{i}^{min} ) ^{2} 
            \\int_{0}^{\\theta_{i}^{min}} \\delta(\\theta-\\theta_{i}) d \\theta
            +
            ( \\theta_{i} - \\theta_{i}^{max} ) ^{2} 
            \\int_{\\theta_{i}^{max}}^{\\pi} \\delta(\\theta-\\theta_{i}) d \\theta
                               
        Where:\n
        :math:`C` is the total number of defined improper angles constraints. \n
        :math:`\\theta_{i}^{min}` is the improper angle constraint lower limit set for constraint i. \n
        :math:`\\theta_{i}^{max}` is the improper angle constraint upper limit set for constraint i. \n
        :math:`\\theta_{i}` is the improper angle computed for constraint i. \n
        :math:`\\delta` is the Dirac delta function. \n
        :math:`\\int_{0}^{\\theta_{i}^{min}} \\delta(\\theta-\\theta_{i}) d \\theta` 
        is equal to 1 if :math:`0 \\leqslant \\theta_{i} \\leqslant \\theta_{i}^{min}` and 0 elsewhere.\n
        :math:`\\int_{\\theta_{i}^{max}}^{\\pi} \\delta(\\theta-\\theta_{i}) d \\theta` 
        is equal to 1 if :math:`\\theta_{i}^{max} \\leqslant \\theta_{i} \\leqslant \\pi` and 0 elsewhere.\n

        :Parameters:
            #. data (numpy.array): The constraint value data to compute standardError.
            
        :Returns:
            #. standardError (number): The calculated standardError of the constraint.
        """
        standardError = 0
        for idx, angle in data.items():
            standardError +=  np.sum(angle["reducedAngles"]**2)
        return FLOAT_TYPE( standardError )

    def get_constraint_value(self):
        """
        Compute all partial Mean Pair Distances (MPD) below the defined minimum distance. 
        
        :Returns:
            #. MPD (dictionary): The MPD dictionary, where keys are the element wise intra and inter molecular MPDs and values are the computed MPDs.
        """
        return self.data
        
    def compute_data(self):
        """ Compute data and update engine constraintsData dictionary. """
        # get angles dictionary slice
        anglesIndexes = []
        for idx in self.engine.pdb.indexes:
            anglesIndexes.extend( self.__atomsLUAD[idx] )
        anglesDict = {}
        for idx in set(anglesIndexes):
            anglesDict[idx] = self.angles[idx] 
        # compute data before move
        dataDict = full_improper_angles_coords( anglesDict         = anglesDict ,
                                                boxCoords          = self.engine.boxCoordinates,
                                                basis              = self.engine.basisVectors ,
                                                isPBC              = self.engine.isPBC,
                                                reduceAngleToUpper = False,
                                                reduceAngleToLower = False,
                                                ncores             = INT_TYPE(1) )                 
        self.set_data( dataDict )
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # set standardError
        self.set_standard_error( self.compute_standard_error(data = dataDict) )
        
    def compute_before_move(self, indexes):
        """ 
        Compute constraint before move is executed
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        # get angles dictionary slice
        anglesIndexes = []
        for idx in indexes:
            anglesIndexes.extend( self.__atomsLUAD[idx] )
        anglesDict = {}
        for idx in set(anglesIndexes):
            anglesDict[idx] = self.angles[idx] 
        # compute data before move
        dataDict = full_improper_angles_coords( anglesDict         = anglesDict ,
                                                boxCoords          = self.engine.boxCoordinates,
                                                basis              = self.engine.basisVectors ,
                                                isPBC              = self.engine.isPBC,
                                                reduceAngleToUpper = False,
                                                reduceAngleToLower = False,
                                                ncores             = INT_TYPE(1) )
        # set data before move
        self.set_active_atoms_data_before_move( dataDict )
        self.set_active_atoms_data_after_move(None)
        
    def compute_after_move(self, indexes, movedBoxCoordinates):
        """ 
        Compute constraint after move is executed
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to.
            #. movedBoxCoordinates (numpy.ndarray): The moved atoms new coordinates.
        """
        # get angles dictionary slice
        anglesIndexes = []
        for idx in indexes:
            anglesIndexes.extend( self.__atomsLUAD[idx] )
        anglesDict = {}
        for idx in set(anglesIndexes):
            anglesDict[idx] = self.__angles[idx] 
        # change coordinates temporarily
        boxData = np.array(self.engine.boxCoordinates[indexes], dtype=FLOAT_TYPE)
        self.engine.boxCoordinates[indexes] = movedBoxCoordinates
        # compute data before move
        dataDict = full_improper_angles_coords( anglesDict         = anglesDict ,
                                                boxCoords          = self.engine.boxCoordinates,
                                                basis              = self.engine.basisVectors ,
                                                isPBC              = self.engine.isPBC,
                                                reduceAngleToUpper = False,
                                                reduceAngleToLower = False,
                                                ncores             = INT_TYPE(1) )
        # set data after move
        self.set_active_atoms_data_after_move( dataDict )
        # reset coordinates
        self.engine.boxCoordinates[indexes] = boxData
  
    def accept_move(self, indexes):
        """ 
        Accept move
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        # get indexes
        anglesIndexes = []
        for idx in indexes:
            anglesIndexes.extend( self.__atomsLUAD[idx] )
        for idx in set(anglesIndexes):
            self.data[idx]["angles"]        = self.activeAtomsDataAfterMove[idx]["angles"]
            self.data[idx]["reducedAngles"] = self.activeAtomsDataAfterMove[idx]["reducedAngles"]
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)

    def reject_move(self, indexes):
        """ 
        Reject move
        
        :Parameters:
            #. indexes (numpy.ndarray): Group atoms indexes the move will be applied to
        """
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)



    
    
            